# -*- coding: utf-8 -*-
# fields.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Fields for Mailbox and Message.
"""
from leap.mail.imap.parser import MBoxParser


class WithMsgFields(object):
    """
    Container class for class-attributes to be shared by
    several message-related classes.
    """
    # indexing
    CONTENT_HASH_KEY = "chash"
    PAYLOAD_HASH_KEY = "phash"

    # Internal representation of Message

    # flags doc
    UID_KEY = "uid"
    MBOX_KEY = "mbox"
    SEEN_KEY = "seen"
    DEL_KEY = "deleted"
    RECENT_KEY = "recent"
    FLAGS_KEY = "flags"
    MULTIPART_KEY = "multi"
    SIZE_KEY = "size"

    # headers
    HEADERS_KEY = "headers"
    DATE_KEY = "date"
    SUBJECT_KEY = "subject"
    PARTS_MAP_KEY = "part_map"
    BODY_KEY = "body"  # link to phash of body
    MSGID_KEY = "msgid"

    # content
    LINKED_FROM_KEY = "lkf"  # XXX not implemented yet!
    RAW_KEY = "raw"
    CTYPE_KEY = "ctype"

    # Mailbox specific keys
    CLOSED_KEY = "closed"
    CREATED_KEY = "created"
    SUBSCRIBED_KEY = "subscribed"
    RW_KEY = "rw"
    LAST_UID_KEY = "lastuid"
    RECENTFLAGS_KEY = "rct"
    HDOCS_SET_KEY = "hdocset"

    # Document Type, for indexing
    TYPE_KEY = "type"
    TYPE_MBOX_VAL = "mbox"
    TYPE_FLAGS_VAL = "flags"
    TYPE_HEADERS_VAL = "head"
    TYPE_CONTENT_VAL = "cnt"
    TYPE_RECENT_VAL = "rct"
    TYPE_HDOCS_SET_VAL = "hdocset"

    INBOX_VAL = "inbox"

    # Flags in Mailbox and Message
    SEEN_FLAG = "\\Seen"
    RECENT_FLAG = "\\Recent"
    ANSWERED_FLAG = "\\Answered"
    FLAGGED_FLAG = "\\Flagged"  # yo dawg
    DELETED_FLAG = "\\Deleted"
    DRAFT_FLAG = "\\Draft"
    NOSELECT_FLAG = "\\Noselect"
    LIST_FLAG = "List"  # is this OK? (no \. ie, no system flag)

    # Fields in mail object
    SUBJECT_FIELD = "Subject"
    DATE_FIELD = "Date"

    # Index  types
    # --------------

    TYPE_IDX = 'by-type'
    TYPE_MBOX_IDX = 'by-type-and-mbox'
    TYPE_MBOX_UID_IDX = 'by-type-and-mbox-and-uid'
    TYPE_SUBS_IDX = 'by-type-and-subscribed'
    TYPE_MSGID_IDX = 'by-type-and-message-id'
    TYPE_MBOX_SEEN_IDX = 'by-type-and-mbox-and-seen'
    TYPE_MBOX_RECT_IDX = 'by-type-and-mbox-and-recent'
    TYPE_MBOX_DEL_IDX = 'by-type-and-mbox-and-deleted'
    TYPE_MBOX_C_HASH_IDX = 'by-type-and-mbox-and-contenthash'
    TYPE_C_HASH_IDX = 'by-type-and-contenthash'
    TYPE_C_HASH_PART_IDX = 'by-type-and-contenthash-and-partnumber'
    TYPE_P_HASH_IDX = 'by-type-and-payloadhash'

    # Tomas created the `recent and seen index`, but the semantic is not too
    # correct since the recent flag is volatile.
    TYPE_MBOX_RECT_SEEN_IDX = 'by-type-and-mbox-and-recent-and-seen'

    # Soledad index for incoming mail, without decrypting errors.
    JUST_MAIL_IDX = "just-mail"
    # XXX the backward-compatible index, will be deprecated at 0.7
    JUST_MAIL_COMPAT_IDX = "just-mail-compat"

    INCOMING_KEY = "incoming"
    ERROR_DECRYPTING_KEY = "errdecr"

    KTYPE = TYPE_KEY
    MBOX_VAL = TYPE_MBOX_VAL
    CHASH_VAL = CONTENT_HASH_KEY
    PHASH_VAL = PAYLOAD_HASH_KEY

    INDEXES = {
        # generic
        TYPE_IDX: [KTYPE],
        TYPE_MBOX_IDX: [KTYPE, MBOX_VAL],
        TYPE_MBOX_UID_IDX: [KTYPE, MBOX_VAL, UID_KEY],

        # mailboxes
        TYPE_SUBS_IDX: [KTYPE, 'bool(subscribed)'],

        # fdocs uniqueness
        TYPE_MBOX_C_HASH_IDX: [KTYPE, MBOX_VAL, CHASH_VAL],

        # headers doc - search by msgid.
        TYPE_MSGID_IDX: [KTYPE, MSGID_KEY],

        # content, headers doc
        TYPE_C_HASH_IDX: [KTYPE, CHASH_VAL],

        # attachment payload dedup
        TYPE_P_HASH_IDX: [KTYPE, PHASH_VAL],

        # messages
        TYPE_MBOX_SEEN_IDX: [KTYPE, MBOX_VAL, 'bool(seen)'],
        TYPE_MBOX_RECT_IDX: [KTYPE, MBOX_VAL, 'bool(recent)'],
        TYPE_MBOX_DEL_IDX: [KTYPE, MBOX_VAL, 'bool(deleted)'],
        TYPE_MBOX_RECT_SEEN_IDX: [KTYPE, MBOX_VAL,
                                  'bool(recent)', 'bool(seen)'],

        # incoming queue
        JUST_MAIL_IDX: [INCOMING_KEY,
                        "bool(%s)" % (ERROR_DECRYPTING_KEY,)],

        # the backward-compatible index, will be deprecated at 0.7
        JUST_MAIL_COMPAT_IDX: [INCOMING_KEY],
    }

    MBOX_KEY = MBOX_VAL

    EMPTY_MBOX = {
        TYPE_KEY: MBOX_KEY,
        TYPE_MBOX_VAL: MBoxParser.INBOX_NAME,
        SUBJECT_KEY: "",
        FLAGS_KEY: [],
        CLOSED_KEY: False,
        SUBSCRIBED_KEY: False,
        RW_KEY: 1,
        LAST_UID_KEY: 0
    }

fields = WithMsgFields  # alias for convenience



# -*- coding: utf-8 -*-
# memorystore.py
# Copyright (C) 2014 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
In-memory transient store for a LEAPIMAPServer.
"""
import contextlib
import logging
import threading
import weakref

from collections import defaultdict
from copy import copy

from enum import Enum
from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from twisted.python import log
from zope.interface import implements

from leap.common.check import leap_assert_type
from leap.mail import size
from leap.mail.utils import empty, phash_iter
from leap.mail.messageflow import MessageProducer
from leap.mail.imap import interfaces
from leap.mail.imap.fields import fields
from leap.mail.imap.messageparts import MessagePartType, MessagePartDoc
from leap.mail.imap.messageparts import RecentFlagsDoc
from leap.mail.imap.messageparts import MessageWrapper
from leap.mail.imap.messageparts import ReferenciableDict

from leap.mail.decorators import deferred_to_thread

logger = logging.getLogger(__name__)


# The default period to do writebacks to the permanent
# soledad storage, in seconds.
SOLEDAD_WRITE_PERIOD = 15

FDOC = MessagePartType.fdoc.key
HDOC = MessagePartType.hdoc.key
CDOCS = MessagePartType.cdocs.key
DOCS_ID = MessagePartType.docs_id.key


@contextlib.contextmanager
def set_bool_flag(obj, att):
    """
    Set a boolean flag to True while we're doing our thing.
    Just to let the world know.
    """
    setattr(obj, att, True)
    try:
        yield True
    except RuntimeError as exc:
        logger.exception(exc)
    finally:
        setattr(obj, att, False)


DirtyState = Enum("none", "dirty", "new")


class MemoryStore(object):
    """
    An in-memory store to where we can write the different parts that
    we split the messages into and buffer them until we write them to the
    permanent storage.

    It uses MessageWrapper instances to represent the message-parts, which are
    indexed by mailbox name and UID.

    It also can be passed a permanent storage as a paremeter (any implementor
    of IMessageStore, in this case a SoledadStore). In this case, a periodic
    dump of the messages stored in memory will be done. The period of the
    writes to the permanent storage is controled by the write_period parameter
    in the constructor.
    """
    implements(interfaces.IMessageStore,
               interfaces.IMessageStoreWriter)

    # TODO We will want to index by chash when we transition to local-only
    # UIDs.

    WRITING_FLAG = "_writing"
    _last_uid_lock = threading.Lock()
    _fdoc_docid_lock = threading.Lock()

    def __init__(self, permanent_store=None,
                 write_period=SOLEDAD_WRITE_PERIOD):
        """
        Initialize a MemoryStore.

        :param permanent_store: a IMessageStore implementor to dump
                                messages to.
        :type permanent_store: IMessageStore
        :param write_period: the interval to dump messages to disk, in seconds.
        :type write_period: int
        """
        self.reactor = reactor

        self._permanent_store = permanent_store
        self._write_period = write_period

        if permanent_store is None:
            self._mbox_closed = defaultdict(lambda: False)

        # Internal Storage: messages
        """
        flags document store.
        _fdoc_store[mbox][uid] = { 'content': 'aaa' }
        """
        self._fdoc_store = defaultdict(lambda: defaultdict(
            lambda: ReferenciableDict({})))

        # Sizes
        """
        {'mbox, uid': <int>}
        """
        self._sizes = {}

        # Internal Storage: payload-hash
        """
        fdocs:doc-id store, stores document IDs for putting
        the dirty flags-docs.
        """
        self._fdoc_id_store = defaultdict(lambda: defaultdict(
            lambda: ''))

        # Internal Storage: content-hash:hdoc
        """
        hdoc-store keeps references to
        the header-documents indexed by content-hash.

        {'chash': { dict-stuff }
        }
        """
        self._hdoc_store = defaultdict(lambda: ReferenciableDict({}))

        # Internal Storage: payload-hash:cdoc
        """
        content-docs stored by payload-hash
        {'phash': { dict-stuff } }
        """
        self._cdoc_store = defaultdict(lambda: ReferenciableDict({}))

        # Internal Storage: content-hash:fdoc
        """
        chash-fdoc-store keeps references to
        the flag-documents indexed by content-hash.

        {'chash': {'mbox-a': weakref.proxy(dict),
                   'mbox-b': weakref.proxy(dict)}
        }
        """
        self._chash_fdoc_store = defaultdict(lambda: defaultdict(lambda: None))

        # Internal Storage: recent-flags store
        """
        recent-flags store keeps one dict per mailbox,
        with the document-id of the u1db document
        and the set of the UIDs that have the recent flag.

        {'mbox-a': {'doc_id': 'deadbeef',
                    'set': {1,2,3,4}
                    }
        }
        """
        # TODO this will have to transition to content-hash
        # indexes after we move to local-only UIDs.

        self._rflags_store = defaultdict(
            lambda: {'doc_id': None, 'set': set([])})

        """
        last-uid store keeps the count of the highest UID
        per mailbox.

        {'mbox-a': 42,
         'mbox-b': 23}
        """
        self._last_uid = defaultdict(lambda: 0)

        """
        known-uids keeps a count of the uids that soledad knows for a given
        mailbox

        {'mbox-a': set([1,2,3])}
        """
        self._known_uids = defaultdict(set)

        """
        mbox-flags is a dict containing flags for each mailbox. this is
        modified from mailbox.getFlags / mailbox.setFlags
        """
        self._mbox_flags = defaultdict(set)

        # New and dirty flags, to set MessageWrapper State.
        self._new = set([])
        self._new_queue = set([])
        self._new_deferreds = {}

        self._dirty = set([])
        self._dirty_queue = set([])
        self._dirty_deferreds = {}

        self._rflags_dirty = set([])

        # Flag for signaling we're busy writing to the disk storage.
        setattr(self, self.WRITING_FLAG, False)

        if self._permanent_store is not None:
            # this producer spits its messages to the permanent store
            # consumer using a queue. We will use that to put
            # our messages to be written.
            self.producer = MessageProducer(permanent_store,
                                            period=0.1)
            # looping call for dumping to SoledadStore
            self._write_loop = LoopingCall(self.write_messages,
                                           permanent_store)

            # We can start the write loop right now, why wait?
            self._start_write_loop()
        else:
            # We have a memory-only store.
            self.producer = None
            self._write_loop = None

    def _start_write_loop(self):
        """
        Start loop for writing to disk database.
        """
        if self._write_loop is None:
            return
        if not self._write_loop.running:
            self._write_loop.start(self._write_period, now=True)

    def _stop_write_loop(self):
        """
        Stop loop for writing to disk database.
        """
        if self._write_loop is None:
            return
        if self._write_loop.running:
            self._write_loop.stop()

    # IMessageStore

    # XXX this would work well for whole message operations.
    # We would have to add a put_flags operation to modify only
    # the flags doc (and set the dirty flag accordingly)

    def create_message(self, mbox, uid, message, observer,
                       notify_on_disk=True):
        """
        Create the passed message into this MemoryStore.

        By default we consider that any message is a new message.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param uid: the UID for the message
        :type uid: int
        :param message: a message to be added
        :type message: MessageWrapper
        :param observer: the deferred that will fire with the
                         UID of the message. If notify_on_disk is True,
                         this will happen when the message is written to
                         Soledad. Otherwise it will fire as soon as we've
                         added the message to the memory store.
        :type observer: Deferred
        :param notify_on_disk: whether the `observer` deferred should
                               wait until the message is written to disk to
                               be fired.
        :type notify_on_disk: bool
        """
        log.msg("Adding new doc to memstore %r (%r)" % (mbox, uid))
        key = mbox, uid

        self._add_message(mbox, uid, message, notify_on_disk)
        self._new.add(key)

        if observer is not None:
            if notify_on_disk:
                # We store this deferred so we can keep track of the pending
                # operations internally.
                # TODO this should fire with the UID !!! -- change that in
                # the soledad store code.
                self._new_deferreds[key] = observer

            else:
                # Caller does not care, just fired and forgot, so we pass
                # a defer that will inmediately have its callback triggered.
                self.reactor.callFromThread(observer.callback, uid)

    def put_message(self, mbox, uid, message, notify_on_disk=True):
        """
        Put an existing message.

        This will also set the dirty flag on the MemoryStore.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param uid: the UID for the message
        :type uid: int
        :param message: a message to be added
        :type message: MessageWrapper
        :param notify_on_disk: whether the deferred that is returned should
                               wait until the message is written to disk to
                               be fired.
        :type notify_on_disk: bool

        :return: a Deferred. if notify_on_disk is True, will be fired
                 when written to the db on disk.
                 Otherwise will fire inmediately
        :rtype: Deferred
        """
        key = mbox, uid
        d = defer.Deferred()
        d.addCallback(lambda result: log.msg("message PUT save: %s" % result))

        self._dirty.add(key)
        self._dirty_deferreds[key] = d
        self._add_message(mbox, uid, message, notify_on_disk)
        return d

    def _add_message(self, mbox, uid, message, notify_on_disk=True):
        """
        Helper method, called by both create_message and put_message.
        See those for parameter documentation.
        """
        msg_dict = message.as_dict()

        fdoc = msg_dict.get(FDOC, None)
        if fdoc is not None:
            fdoc_store = self._fdoc_store[mbox][uid]
            fdoc_store.update(fdoc)
            chash_fdoc_store = self._chash_fdoc_store

            # content-hash indexing
            chash = fdoc.get(fields.CONTENT_HASH_KEY)
            chash_fdoc_store[chash][mbox] = weakref.proxy(
                self._fdoc_store[mbox][uid])

        hdoc = msg_dict.get(HDOC, None)
        if hdoc is not None:
            chash = hdoc.get(fields.CONTENT_HASH_KEY)
            hdoc_store = self._hdoc_store[chash]
            hdoc_store.update(hdoc)

        cdocs = message.cdocs
        for cdoc in cdocs.values():
            phash = cdoc.get(fields.PAYLOAD_HASH_KEY, None)
            if not phash:
                continue
            cdoc_store = self._cdoc_store[phash]
            cdoc_store.update(cdoc)

        # Update memory store size
        # XXX this should use [mbox][uid]
        # TODO --- this has to be deferred to thread,
        # TODO add hdoc and cdocs sizes too
        # it's slowing things down here.
        # key = mbox, uid
        # self._sizes[key] = size.get_size(self._fdoc_store[key])

    def purge_fdoc_store(self, mbox):
        """
        Purge the empty documents from a fdoc store.
        Called during initialization of the SoledadMailbox

        :param mbox: the mailbox
        :type mbox: str or unicode
        """
        # XXX This is really a workaround until I find the conditions
        # that are making the empty items remain there.
        # This happens, for instance, after running several times
        # the regression test, that issues a store deleted + expunge + select
        # The items are being correclty deleted, but in succesive appends
        # the empty items with previously deleted uids reappear as empty
        # documents. I suspect it's a timing condition with a previously
        # evaluated sequence being used after the items has been removed.

        for uid, value in self._fdoc_store[mbox].items():
            if empty(value):
                del self._fdoc_store[mbox][uid]

    def get_docid_for_fdoc(self, mbox, uid):
        """
        Return Soledad document id for the flags-doc for a given mbox and uid,
        or None of no flags document could be found.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param uid: the message UID
        :type uid: int
        :rtype: unicode or None
        """
        with self._fdoc_docid_lock:
            doc_id = self._fdoc_id_store[mbox][uid]

        if empty(doc_id):
            fdoc = self._permanent_store.get_flags_doc(mbox, uid)
            if empty(fdoc) or empty(fdoc.content):
                return None
            doc_id = fdoc.doc_id
            self._fdoc_id_store[mbox][uid] = doc_id

        return doc_id

    def get_message(self, mbox, uid, dirtystate=DirtyState.none,
                    flags_only=False):
        """
        Get a MessageWrapper for the given mbox and uid combination.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param uid: the message UID
        :type uid: int
        :param dirtystate: DirtyState enum: one of `dirty`, `new`
                           or `none` (default)
        :type dirtystate: enum
        :param flags_only: whether the message should carry only a reference
                           to the flags document.
        :type flags_only: bool
        :

        :return: MessageWrapper or None
        """
        if dirtystate == DirtyState.dirty:
            flags_only = True

        key = mbox, uid

        fdoc = self._fdoc_store[mbox][uid]
        if empty(fdoc):
            return None

        new, dirty = False, False
        if dirtystate == DirtyState.none:
            new, dirty = self._get_new_dirty_state(key)
        if dirtystate == DirtyState.dirty:
            new, dirty = False, True
        if dirtystate == DirtyState.new:
            new, dirty = True, False

        if flags_only:
            return MessageWrapper(fdoc=fdoc,
                                  new=new, dirty=dirty,
                                  memstore=weakref.proxy(self))
        else:
            chash = fdoc.get(fields.CONTENT_HASH_KEY)
            hdoc = self._hdoc_store[chash]
            if empty(hdoc):
                hdoc = self._permanent_store.get_headers_doc(chash)
                if empty(hdoc):
                    return None
                if not empty(hdoc.content):
                    self._hdoc_store[chash] = hdoc.content
                    hdoc = hdoc.content
            cdocs = None

            pmap = hdoc.get(fields.PARTS_MAP_KEY, None)
            if new and pmap is not None:
                # take the different cdocs for write...
                cdoc_store = self._cdoc_store
                cdocs_list = phash_iter(hdoc)
                cdocs = dict(enumerate(
                    [cdoc_store[phash] for phash in cdocs_list], 1))

            return MessageWrapper(fdoc=fdoc, hdoc=hdoc, cdocs=cdocs,
                                  new=new, dirty=dirty,
                                  memstore=weakref.proxy(self))

    def remove_message(self, mbox, uid):
        """
        Remove a Message from this MemoryStore.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param uid: the message UID
        :type uid: int
        """
        # XXX For the moment we are only removing the flags and headers
        # docs. The rest we leave there polluting your hard disk,
        # until we think about a good way of deorphaning.

        # XXX implement elijah's idea of using a PUT document as a
        # token to ensure consistency in the removal.

        try:
            del self._fdoc_store[mbox][uid]
        except KeyError:
            pass

        try:
            key = mbox, uid
            self._new.discard(key)
            self._dirty.discard(key)
            if key in self._sizes:
                del self._sizes[key]
            self._known_uids[mbox].discard(uid)
        except KeyError:
            pass
        except Exception as exc:
            logger.error("error while removing message!")
            logger.exception(exc)
        try:
            with self._fdoc_docid_lock:
                del self._fdoc_id_store[mbox][uid]
        except KeyError:
            pass
        except Exception as exc:
            logger.error("error while removing message!")
            logger.exception(exc)

    # IMessageStoreWriter

    @deferred_to_thread
    def write_messages(self, store):
        """
        Write the message documents in this MemoryStore to a different store.

        :param store: the IMessageStore to write to
        :rtype: False if queue is not empty, None otherwise.
        """
        # For now, we pass if the queue is not empty, to avoid duplicate
        # queuing.
        # We would better use a flag to know when we've already enqueued an
        # item.

        # XXX this could return the deferred for all the enqueued operations

        if not self.producer.is_queue_empty():
            return False

        if any(map(lambda i: not empty(i), (self._new, self._dirty))):
            logger.info("Writing messages to Soledad...")

        # TODO change for lock, and make the property access
        # is accquired
        with set_bool_flag(self, self.WRITING_FLAG):
            for rflags_doc_wrapper in self.all_rdocs_iter():
                self.producer.push(rflags_doc_wrapper,
                                   state=self.producer.STATE_DIRTY)
            for msg_wrapper in self.all_new_msg_iter():
                self.producer.push(msg_wrapper,
                                   state=self.producer.STATE_NEW)
            for msg_wrapper in self.all_dirty_msg_iter():
                self.producer.push(msg_wrapper,
                                   state=self.producer.STATE_DIRTY)

    # MemoryStore specific methods.

    def get_uids(self, mbox):
        """
        Get all uids for a given mbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :rtype: list
        """
        return self._fdoc_store[mbox].keys()

    def get_soledad_known_uids(self, mbox):
        """
        Get all uids that soledad knows about, from the memory cache.
        :param mbox: the mailbox
        :type mbox: str or unicode
        :rtype: list
        """
        return self._known_uids.get(mbox, [])

    # last_uid

    def get_last_uid(self, mbox):
        """
        Return the highest UID for a given mbox.
        It will be the highest between the highest uid in the message store for
        the mailbox, and the soledad integer cache.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :rtype: int
        """
        uids = self.get_uids(mbox)
        last_mem_uid = uids and max(uids) or 0
        last_soledad_uid = self.get_last_soledad_uid(mbox)
        return max(last_mem_uid, last_soledad_uid)

    def get_last_soledad_uid(self, mbox):
        """
        Get last uid for a given mbox from the soledad integer cache.

        :param mbox: the mailbox
        :type mbox: str or unicode
        """
        return self._last_uid.get(mbox, 0)

    def set_last_soledad_uid(self, mbox, value):
        """
        Set last uid for a given mbox in the soledad integer cache.
        SoledadMailbox should prime this value during initialization.
        Other methods (during message adding) SHOULD call
        `increment_last_soledad_uid` instead.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param value: the value to set
        :type value: int
        """
        # can be long???
        # leap_assert_type(value, int)
        logger.info("setting last soledad uid for %s to %s" %
                    (mbox, value))
        # if we already have a value here, don't do anything
        with self._last_uid_lock:
            if not self._last_uid.get(mbox, None):
                self._last_uid[mbox] = value

    def set_known_uids(self, mbox, value):
        """
        Set the value fo the known-uids set for this mbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param value: a sequence of integers to be added to the set.
        :type value: tuple
        """
        current = self._known_uids[mbox]
        self._known_uids[mbox] = current.union(set(value))

    def increment_last_soledad_uid(self, mbox):
        """
        Increment by one the soledad integer cache for the last_uid for
        this mbox, and fire a defer-to-thread to update the soledad value.
        The caller should lock the call tho this method.

        :param mbox: the mailbox
        :type mbox: str or unicode
        """
        with self._last_uid_lock:
            self._last_uid[mbox] += 1
            value = self._last_uid[mbox]
            self.reactor.callInThread(self.write_last_uid, mbox, value)
            return value

    def write_last_uid(self, mbox, value):
        """
        Increment the soledad integer cache for the highest uid value.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param value: the value to set
        :type value: int
        """
        leap_assert_type(value, int)
        if self._permanent_store:
            self._permanent_store.write_last_uid(mbox, value)

    def load_flag_docs(self, mbox, flag_docs):
        """
        Load the flag documents for the given mbox.
        Used during initial flag docs prefetch.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param flag_docs: a dict with the content for the flag docs, indexed
                          by uid.
        :type flag_docs: dict
        """
        # We can do direct assignments cause we know this will only
        # be called during initialization of the mailbox.
        # TODO could hook here a sanity-check
        # for duplicates

        fdoc_store = self._fdoc_store[mbox]
        chash_fdoc_store = self._chash_fdoc_store
        for uid in flag_docs:
            rdict = ReferenciableDict(flag_docs[uid])
            fdoc_store[uid] = rdict
            # populate chash dict too, to avoid fdoc duplication
            chash = flag_docs[uid]["chash"]
            chash_fdoc_store[chash][mbox] = weakref.proxy(
                self._fdoc_store[mbox][uid])

    def update_flags(self, mbox, uid, fdoc):
        """
        Update the flag document for a given mbox and uid combination,
        and set the dirty flag.
        We could use put_message, but this is faster.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param uid: the uid of the message
        :type uid: int

        :param fdoc: a dict with the content for the flag docs
        :type fdoc: dict
        """
        key = mbox, uid
        self._fdoc_store[mbox][uid].update(fdoc)
        self._dirty.add(key)

    def load_header_docs(self, header_docs):
        """
        Load the flag documents for the given mbox.
        Used during header docs prefetch, and during cache after
        a read from soledad if the hdoc property in message did not
        find its value in here.

        :param flag_docs: a dict with the content for the flag docs.
        :type flag_docs: dict
        """
        hdoc_store = self._hdoc_store
        for chash in header_docs:
            hdoc_store[chash] = ReferenciableDict(header_docs[chash])

    def all_flags(self, mbox):
        """
        Return a dictionary with all the flags for a given mbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :rtype: dict
        """
        fdict = {}
        uids = self.get_uids(mbox)
        fstore = self._fdoc_store[mbox]

        for uid in uids:
            try:
                fdict[uid] = fstore[uid][fields.FLAGS_KEY]
            except KeyError:
                continue
        return fdict

    def all_headers(self, mbox):
        """
        Return a dictionary with all the header docs for a given mbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :rtype: dict
        """
        headers_dict = {}
        uids = self.get_uids(mbox)
        fdoc_store = self._fdoc_store[mbox]
        hdoc_store = self._hdoc_store

        for uid in uids:
            try:
                chash = fdoc_store[uid][fields.CONTENT_HASH_KEY]
                hdoc = hdoc_store[chash]
                if not empty(hdoc):
                    headers_dict[uid] = hdoc
            except KeyError:
                continue
        return headers_dict

    # Counting sheeps...

    def count_new_mbox(self, mbox):
        """
        Count the new messages by mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :return: number of new messages
        :rtype: int
        """
        return len([(m, uid) for m, uid in self._new if mbox == mbox])

    # XXX used at all?
    def count_new(self):
        """
        Count all the new messages in the MemoryStore.

        :rtype: int
        """
        return len(self._new)

    def count(self, mbox):
        """
        Return the count of messages for a given mbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :return: number of messages
        :rtype: int
        """
        return len(self._fdoc_store[mbox])

    def unseen_iter(self, mbox):
        """
        Get an iterator for the message UIDs with no `seen` flag
        for a given mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :return: iterator through unseen message doc UIDs
        :rtype: iterable
        """
        fdocs = self._fdoc_store[mbox]

        return [uid for uid, value
                in fdocs.items()
                if fields.SEEN_FLAG not in value.get(fields.FLAGS_KEY, [])]

    def get_cdoc_from_phash(self, phash):
        """
        Return a content-document by its payload-hash.

        :param phash: the payload hash to check against
        :type phash: str or unicode
        :rtype: MessagePartDoc
        """
        doc = self._cdoc_store.get(phash, None)

        # XXX return None for consistency?

        # XXX have to keep a mapping between phash and its linkage
        # info, to know if this payload is been already saved or not.
        # We will be able to get this from the linkage-docs,
        # not yet implemented.
        new = True
        dirty = False
        return MessagePartDoc(
            new=new, dirty=dirty, store="mem",
            part=MessagePartType.cdoc,
            content=doc,
            doc_id=None)

    def get_fdoc_from_chash(self, chash, mbox):
        """
        Return a flags-document by its content-hash and a given mailbox.
        Used during content-duplication detection while copying or adding a
        message.

        :param chash: the content hash to check against
        :type chash: str or unicode
        :param mbox: the mailbox
        :type mbox: str or unicode

        :return: MessagePartDoc. It will return None if the flags document
                 has empty content or it is flagged as \\Deleted.
        """
        fdoc = self._chash_fdoc_store[chash][mbox]

        # a couple of special cases.
        # 1. We might have a doc with empty content...
        if empty(fdoc):
            return None

        # 2. ...Or the message could exist, but being flagged for deletion.
        # We want to create a new one in this case.
        # Hmmm what if the deletion is un-done?? We would end with a
        # duplicate...
        if fdoc and fields.DELETED_FLAG in fdoc.get(fields.FLAGS_KEY, []):
            return None

        uid = fdoc[fields.UID_KEY]
        key = mbox, uid
        new = key in self._new
        dirty = key in self._dirty

        return MessagePartDoc(
            new=new, dirty=dirty, store="mem",
            part=MessagePartType.fdoc,
            content=fdoc,
            doc_id=None)

    def iter_fdoc_keys(self):
        """
        Return a generator through all the mbox, uid keys in the flags-doc
        store.
        """
        fdoc_store = self._fdoc_store
        for mbox in fdoc_store:
            for uid in fdoc_store[mbox]:
                yield mbox, uid

    def all_new_msg_iter(self):
        """
        Return generator that iterates through all new messages.

        :return: generator of MessageWrappers
        :rtype: generator
        """
        gm = self.get_message
        # need to freeze, set can change during iteration
        new = [gm(*key, dirtystate=DirtyState.new) for key in tuple(self._new)]
        # move content from new set to the queue
        self._new_queue.update(self._new)
        self._new.difference_update(self._new)
        return new

    def all_dirty_msg_iter(self):
        """
        Return generator that iterates through all dirty messages.

        :return: generator of MessageWrappers
        :rtype: generator
        """
        gm = self.get_message
        # need to freeze, set can change during iteration
        dirty = [gm(*key, flags_only=True, dirtystate=DirtyState.dirty)
                 for key in tuple(self._dirty)]
        # move content from new and dirty sets to the queue

        self._dirty_queue.update(self._dirty)
        self._dirty.difference_update(self._dirty)
        return dirty

    def all_deleted_uid_iter(self, mbox):
        """
        Return a list with the UIDs for all messags
        with deleted flag in a given mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :return: list of integers
        :rtype: list
        """
        # This *needs* to return a fixed sequence. Otherwise the dictionary len
        # will change during iteration, when we modify it
        fdocs = self._fdoc_store[mbox]
        return [uid for uid, value
                in fdocs.items()
                if fields.DELETED_FLAG in value.get(fields.FLAGS_KEY, [])]

    # new, dirty flags

    def _get_new_dirty_state(self, key):
        """
        Return `new` and `dirty` flags for a given message.

        :param key: the key for the message, in the form mbox, uid
        :type key: tuple
        :return: tuple of bools
        :rtype: tuple
        """
        # TODO change indexing of sets to [mbox][key] too.
        # XXX should return *first* the news, and *then* the dirty...

        # TODO should query in queues too , true?
        #
        return map(lambda _set: key in _set, (self._new, self._dirty))

    def set_new_queued(self, key):
        """
        Add the key value to the `new-queue` set.

        :param key: the key for the message, in the form mbox, uid
        :type key: tuple
        """
        self._new_queue.add(key)

    def unset_new_queued(self, key):
        """
        Remove the key value from the `new-queue` set.

        :param key: the key for the message, in the form mbox, uid
        :type key: tuple
        """
        self._new_queue.discard(key)
        deferreds = self._new_deferreds
        d = deferreds.get(key, None)
        if d:
            # XXX use a namedtuple for passing the result
            # when we check it in the other side.
            d.callback('%s, ok' % str(key))
            deferreds.pop(key)

    def set_dirty_queued(self, key):
        """
        Add the key value to the `dirty-queue` set.

        :param key: the key for the message, in the form mbox, uid
        :type key: tuple
        """
        self._dirty_queue.add(key)

    def unset_dirty_queued(self, key):
        """
        Remove the key value from the `dirty-queue` set.

        :param key: the key for the message, in the form mbox, uid
        :type key: tuple
        """
        self._dirty_queue.discard(key)
        deferreds = self._dirty_deferreds
        d = deferreds.get(key, None)
        if d:
            # XXX use a namedtuple for passing the result
            # when we check it in the other side.
            d.callback('%s, ok' % str(key))
            deferreds.pop(key)

    # Recent Flags

    def set_recent_flag(self, mbox, uid):
        """
        Set the `Recent` flag for a given mailbox and UID.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param uid: the message UID
        :type uid: int
        """
        self._rflags_dirty.add(mbox)
        self._rflags_store[mbox]['set'].add(uid)

    # TODO --- nice but unused
    def unset_recent_flag(self, mbox, uid):
        """
        Unset the `Recent` flag for a given mailbox and UID.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param uid: the message UID
        :type uid: int
        """
        self._rflags_store[mbox]['set'].discard(uid)

    def set_recent_flags(self, mbox, value):
        """
        Set the value for the set of the recent flags.
        Used from the property in the MessageCollection.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param value: a sequence of flags to set
        :type value: sequence
        """
        self._rflags_dirty.add(mbox)
        self._rflags_store[mbox]['set'] = set(value)

    def load_recent_flags(self, mbox, flags_doc):
        """
        Load the passed flags document in the recent flags store, for a given
        mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param flags_doc: A dictionary containing the `doc_id` of the Soledad
                          flags-document for this mailbox, and the `set`
                          of uids marked with that flag.
        """
        self._rflags_store[mbox] = flags_doc

    def get_recent_flags(self, mbox):
        """
        Return the set of UIDs with the `Recent` flag for this mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :rtype: set, or None
        """
        rflag_for_mbox = self._rflags_store.get(mbox, None)
        if not rflag_for_mbox:
            return None
        return self._rflags_store[mbox]['set']

    def all_rdocs_iter(self):
        """
        Return an iterator through all in-memory recent flag dicts, wrapped
        under a RecentFlagsDoc namedtuple.
        Used for saving to disk.

        :return: a generator of RecentFlagDoc
        :rtype: generator
        """
        # XXX use enums
        DOC_ID = "doc_id"
        SET = "set"

        rflags_store = self._rflags_store

        def get_rdoc(mbox, rdict):
            mbox_rflag_set = rdict[SET]
            recent_set = copy(mbox_rflag_set)
            # zero it!
            mbox_rflag_set.difference_update(mbox_rflag_set)
            return RecentFlagsDoc(
                doc_id=rflags_store[mbox][DOC_ID],
                content={
                    fields.TYPE_KEY: fields.TYPE_RECENT_VAL,
                    fields.MBOX_KEY: mbox,
                    fields.RECENTFLAGS_KEY: list(recent_set)
                })

        return (get_rdoc(mbox, rdict) for mbox, rdict in rflags_store.items()
                if not empty(rdict[SET]))

    # Methods that mirror the IMailbox interface

    def remove_all_deleted(self, mbox):
        """
        Remove all messages flagged \\Deleted from this Memory Store only.
        Called from `expunge`

        :param mbox: the mailbox
        :type mbox: str or unicode
        :return: a list of UIDs
        :rtype: list
        """
        mem_deleted = self.all_deleted_uid_iter(mbox)
        for uid in mem_deleted:
            self.remove_message(mbox, uid)
        return mem_deleted

    def stop_and_flush(self):
        """
        Stop the write loop and trigger a write to the producer.
        """
        self._stop_write_loop()
        if self._permanent_store is not None:
            # XXX we should check if we did get a True value on this
            # operation. If we got False we should retry! (queue was not empty)
            self.write_messages(self._permanent_store)
            self.producer.flush()

    def expunge(self, mbox, observer):
        """
        Remove all messages flagged \\Deleted, from the Memory Store
        and from the permanent store also.

        It first queues up a last write, and wait for the deferreds to be done
        before continuing.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param observer: a deferred that will be fired when expunge is done
        :type observer: Deferred
        """
        soledad_store = self._permanent_store
        if soledad_store is None:
            # just-in memory store, easy then.
            self._delete_from_memory(mbox, observer)
            return

        # We have a soledad storage.
        try:
            # Stop and trigger last write
            self.stop_and_flush()
            # Wait on the writebacks to finish

            # XXX what if pending deferreds is empty?
            pending_deferreds = (self._new_deferreds.get(mbox, []) +
                                 self._dirty_deferreds.get(mbox, []))
            d1 = defer.gatherResults(pending_deferreds, consumeErrors=True)
            d1.addCallback(
                self._delete_from_soledad_and_memory, mbox, observer)
        except Exception as exc:
            logger.exception(exc)

    def _delete_from_memory(self, mbox, observer):
        """
        Remove all messages marked as deleted from soledad and memory.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param observer: a deferred that will be fired when expunge is done
        :type observer: Deferred
        """
        mem_deleted = self.remove_all_deleted(mbox)
        observer.callback(mem_deleted)

    def _delete_from_soledad_and_memory(self, result, mbox, observer):
        """
        Remove all messages marked as deleted from soledad and memory.

        :param result: ignored. the result of the deferredList that triggers
                       this as a callback from `expunge`.
        :param mbox: the mailbox
        :type mbox: str or unicode
        :param observer: a deferred that will be fired when expunge is done
        :type observer: Deferred
        """
        all_deleted = []
        soledad_store = self._permanent_store

        try:
            # 1. Delete all messages marked as deleted in soledad.
            logger.debug("DELETING FROM SOLEDAD ALL FOR %r" % (mbox,))
            sol_deleted = soledad_store.remove_all_deleted(mbox)

            try:
                self._known_uids[mbox].difference_update(set(sol_deleted))
            except Exception as exc:
                logger.exception(exc)

            # 2. Delete all messages marked as deleted in memory.
            logger.debug("DELETING FROM MEM ALL FOR %r" % (mbox,))
            mem_deleted = self.remove_all_deleted(mbox)

            all_deleted = set(mem_deleted).union(set(sol_deleted))
            logger.debug("deleted %r" % all_deleted)
        except Exception as exc:
            logger.exception(exc)
        finally:
            self._start_write_loop()

        observer.callback(all_deleted)

    # Mailbox documents and attributes

    # This could be also be cached in memstore, but proxying directly
    # to soledad since it's not too performance-critical.

    def get_mbox_doc(self, mbox):
        """
        Return the soledad document for a given mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :rtype: SoledadDocument or None.
        """
        if self.permanent_store is not None:
            return self.permanent_store.get_mbox_document(mbox)
        else:
            return None

    def get_mbox_closed(self, mbox):
        """
        Return the closed attribute for a given mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :rtype: bool
        """
        if self.permanent_store is not None:
            return self.permanent_store.get_mbox_closed(mbox)
        else:
            return self._mbox_closed[mbox]

    def set_mbox_closed(self, mbox, closed):
        """
        Set the closed attribute for a given mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        """
        if self.permanent_store is not None:
            self.permanent_store.set_mbox_closed(mbox, closed)
        else:
            self._mbox_closed[mbox] = closed

    def get_mbox_flags(self, mbox):
        """
        Get the flags for a given mbox.
        :rtype: list
        """
        return sorted(self._mbox_flags[mbox])

    def set_mbox_flags(self, mbox, flags):
        """
        Set the mbox flags
        """
        self._mbox_flags[mbox] = set(flags)
        # TODO
        # This should write to the permanent store!!!

    # Rename flag-documents

    def rename_fdocs_mailbox(self, old_mbox, new_mbox):
        """
        Change the mailbox name for all flag documents in a given mailbox.
        Used from account.rename

        :param old_mbox: name for the old mbox
        :type old_mbox: str or unicode
        :param new_mbox: name for the new mbox
        :type new_mbox: str or unicode
        """
        fs = self._fdoc_store
        keys = fs[old_mbox].keys()
        for k in keys:
            fdoc = fs[old_mbox][k]
            fdoc['mbox'] = new_mbox
            fs[new_mbox][k] = fdoc
            fs[old_mbox].pop(k)
            self._dirty.add((new_mbox, k))

    # Dump-to-disk controls.

    @property
    def is_writing(self):
        """
        Property that returns whether the store is currently writing its
        internal state to a permanent storage.

        Used to evaluate whether the CHECK command can inform that the field
        is clear to proceed, or waiting for the write operations to complete
        is needed instead.

        :rtype: bool
        """
        # FIXME this should return a deferred !!!
        # XXX ----- can fire when all new + dirty deferreds
        # are done (gatherResults)
        return getattr(self, self.WRITING_FLAG)

    @property
    def permanent_store(self):
        return self._permanent_store

    # Memory management.

    def get_size(self):
        """
        Return the size of the internal storage.
        Use for calculating the limit beyond which we should flush the store.

        :rtype: int
        """
        return reduce(lambda x, y: x + y, self._sizes, 0)



# -*- coding: utf-8 -*-
# server.py
# Copyright (C) 2014 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Leap IMAP4 Server Implementation.
"""
from copy import copy

from twisted import cred
from twisted.internet.defer import maybeDeferred
from twisted.mail import imap4
from twisted.python import log

from leap.common import events as leap_events
from leap.common.check import leap_assert, leap_assert_type
from leap.common.events.events_pb2 import IMAP_CLIENT_LOGIN
from leap.soledad.client import Soledad

# imports for LITERAL+ patch
from twisted.internet import defer, interfaces
from twisted.mail.imap4 import IllegalClientResponse
from twisted.mail.imap4 import LiteralString, LiteralFile


class LeapIMAPServer(imap4.IMAP4Server):
    """
    An IMAP4 Server with mailboxes backed by soledad
    """
    def __init__(self, *args, **kwargs):
        # pop extraneous arguments
        soledad = kwargs.pop('soledad', None)
        uuid = kwargs.pop('uuid', None)
        userid = kwargs.pop('userid', None)

        leap_assert(soledad, "need a soledad instance")
        leap_assert_type(soledad, Soledad)
        leap_assert(uuid, "need a user in the initialization")

        self._userid = userid

        # initialize imap server!
        imap4.IMAP4Server.__init__(self, *args, **kwargs)

        # we should initialize the account here,
        # but we move it to the factory so we can
        # populate the test account properly (and only once
        # per session)

        from twisted.internet import reactor
        self.reactor = reactor

    def lineReceived(self, line):
        """
        Attempt to parse a single line from the server.

        :param line: the line from the server, without the line delimiter.
        :type line: str
        """
        if self.theAccount.closed is True and self.state != "unauth":
            log.msg("Closing the session. State: unauth")
            self.state = "unauth"

        if "login" in line.lower():
            # avoid to log the pass, even though we are using a dummy auth
            # by now.
            msg = line[:7] + " [...]"
        else:
            msg = copy(line)
        log.msg('rcv (%s): %s' % (self.state, msg))
        imap4.IMAP4Server.lineReceived(self, line)

    def authenticateLogin(self, username, password):
        """
        Lookup the account with the given parameters, and deny
        the improper combinations.

        :param username: the username that is attempting authentication.
        :type username: str
        :param password: the password to authenticate with.
        :type password: str
        """
        # XXX this should use portal:
        # return portal.login(cred.credentials.UsernamePassword(user, pass)
        if username != self._userid:
            # bad username, reject.
            raise cred.error.UnauthorizedLogin()
        # any dummy password is allowed so far. use realm instead!
        leap_events.signal(IMAP_CLIENT_LOGIN, "1")
        return imap4.IAccount, self.theAccount, lambda: None

    def do_FETCH(self, tag, messages, query, uid=0):
        """
        Overwritten fetch dispatcher to use the fast fetch_flags
        method
        """
        if not query:
            self.sendPositiveResponse(tag, 'FETCH complete')
            return

        cbFetch = self._IMAP4Server__cbFetch
        ebFetch = self._IMAP4Server__ebFetch

        if len(query) == 1 and str(query[0]) == "flags":
            self._oldTimeout = self.setTimeout(None)
            # no need to call iter, we get a generator
            maybeDeferred(
                self.mbox.fetch_flags, messages, uid=uid
            ).addCallback(
                cbFetch, tag, query, uid
            ).addErrback(ebFetch, tag)

        elif len(query) == 1 and str(query[0]) == "rfc822.header":
            self._oldTimeout = self.setTimeout(None)
            # no need to call iter, we get a generator
            maybeDeferred(
                self.mbox.fetch_headers, messages, uid=uid
            ).addCallback(
                cbFetch, tag, query, uid
            ).addErrback(ebFetch, tag)
        else:
            self._oldTimeout = self.setTimeout(None)
            # no need to call iter, we get a generator
            maybeDeferred(
                self.mbox.fetch, messages, uid=uid
            ).addCallback(
                cbFetch, tag, query, uid
            ).addErrback(
                ebFetch, tag)

    select_FETCH = (do_FETCH, imap4.IMAP4Server.arg_seqset,
                    imap4.IMAP4Server.arg_fetchatt)

    def notifyNew(self, ignored=None):
        """
        Notify new messages to listeners.
        """
        self.reactor.callFromThread(self.mbox.notify_new)

    def _cbSelectWork(self, mbox, cmdName, tag):
        """
        Callback for selectWork, patched to avoid conformance errors due to
        incomplete UIDVALIDITY line.
        """
        if mbox is None:
            self.sendNegativeResponse(tag, 'No such mailbox')
            return
        if '\\noselect' in [s.lower() for s in mbox.getFlags()]:
            self.sendNegativeResponse(tag, 'Mailbox cannot be selected')
            return

        flags = mbox.getFlags()
        self.sendUntaggedResponse(str(mbox.getMessageCount()) + ' EXISTS')
        self.sendUntaggedResponse(str(mbox.getRecentCount()) + ' RECENT')
        self.sendUntaggedResponse('FLAGS (%s)' % ' '.join(flags))

        # Patched -------------------------------------------------------
        # imaptest was complaining about the incomplete line, we're adding
        # "UIDs valid" here.
        self.sendPositiveResponse(
            None, '[UIDVALIDITY %d] UIDs valid' % mbox.getUIDValidity())
        # ----------------------------------------------------------------

        s = mbox.isWriteable() and 'READ-WRITE' or 'READ-ONLY'
        mbox.addListener(self)
        self.sendPositiveResponse(tag, '[%s] %s successful' % (s, cmdName))
        self.state = 'select'
        self.mbox = mbox

    def checkpoint(self):
        """
        Called when the client issues a CHECK command.

        This should perform any checkpoint operations required by the server.
        It may be a long running operation, but may not block.  If it returns
        a deferred, the client will only be informed of success (or failure)
        when the deferred's callback (or errback) is invoked.
        """
        # TODO return the output of _memstore.is_writing
        # XXX and that should return a deferred!
        return None

    #############################################################
    #
    # Twisted imap4 patch to support LITERAL+ extension
    # TODO send this patch upstream asap!
    #
    #############################################################

    def capabilities(self):
        cap = {'AUTH': self.challengers.keys()}
        if self.ctx and self.canStartTLS:
            t = self.transport
            ti = interfaces.ISSLTransport
            if not self.startedTLS and ti(t, None) is None:
                cap['LOGINDISABLED'] = None
                cap['STARTTLS'] = None
        cap['NAMESPACE'] = None
        cap['IDLE'] = None
        # patched ############
        cap['LITERAL+'] = None
        ######################
        return cap

    def _stringLiteral(self, size, literal_plus=False):
        if size > self._literalStringLimit:
            raise IllegalClientResponse(
                "Literal too long! I accept at most %d octets" %
                (self._literalStringLimit,))
        d = defer.Deferred()
        self.parseState = 'pending'
        self._pendingLiteral = LiteralString(size, d)
        # Patched ###########################################################
        if not literal_plus:
            self.sendContinuationRequest('Ready for %d octets of text' % size)
        #####################################################################
        self.setRawMode()
        return d

    def _fileLiteral(self, size, literal_plus=False):
        d = defer.Deferred()
        self.parseState = 'pending'
        self._pendingLiteral = LiteralFile(size, d)
        if not literal_plus:
            self.sendContinuationRequest('Ready for %d octets of data' % size)
        self.setRawMode()
        return d

    def arg_astring(self, line):
        """
        Parse an astring from the line, return (arg, rest), possibly
        via a deferred (to handle literals)
        """
        line = line.strip()
        if not line:
            raise IllegalClientResponse("Missing argument")
        d = None
        arg, rest = None, None
        if line[0] == '"':
            try:
                spam, arg, rest = line.split('"', 2)
                rest = rest[1:]  # Strip space
            except ValueError:
                raise IllegalClientResponse("Unmatched quotes")
        elif line[0] == '{':
            # literal
            if line[-1] != '}':
                raise IllegalClientResponse("Malformed literal")

            # Patched ################
            if line[-2] == "+":
                literalPlus = True
                size_end = -2
            else:
                literalPlus = False
                size_end = -1

            try:
                size = int(line[1:size_end])
            except ValueError:
                raise IllegalClientResponse(
                    "Bad literal size: " + line[1:size_end])
            d = self._stringLiteral(size, literalPlus)
            ##########################
        else:
            arg = line.split(' ', 1)
            if len(arg) == 1:
                arg.append('')
            arg, rest = arg
        return d or (arg, rest)

    def arg_literal(self, line):
        """
        Parse a literal from the line
        """
        if not line:
            raise IllegalClientResponse("Missing argument")

        if line[0] != '{':
            raise IllegalClientResponse("Missing literal")

        if line[-1] != '}':
            raise IllegalClientResponse("Malformed literal")

        # Patched ##################
        if line[-2] == "+":
            literalPlus = True
            size_end = -2
        else:
            literalPlus = False
            size_end = -1

        try:
            size = int(line[1:size_end])
        except ValueError:
            raise IllegalClientResponse(
                "Bad literal size: " + line[1:size_end])

        return self._fileLiteral(size, literalPlus)
        #############################

    # Need to override the command table after patching
    # arg_astring and arg_literal

    do_LOGIN = imap4.IMAP4Server.do_LOGIN
    do_CREATE = imap4.IMAP4Server.do_CREATE
    do_DELETE = imap4.IMAP4Server.do_DELETE
    do_RENAME = imap4.IMAP4Server.do_RENAME
    do_SUBSCRIBE = imap4.IMAP4Server.do_SUBSCRIBE
    do_UNSUBSCRIBE = imap4.IMAP4Server.do_UNSUBSCRIBE
    do_STATUS = imap4.IMAP4Server.do_STATUS
    do_APPEND = imap4.IMAP4Server.do_APPEND
    do_COPY = imap4.IMAP4Server.do_COPY

    _selectWork = imap4.IMAP4Server._selectWork
    _listWork = imap4.IMAP4Server._listWork
    arg_plist = imap4.IMAP4Server.arg_plist
    arg_seqset = imap4.IMAP4Server.arg_seqset
    opt_plist = imap4.IMAP4Server.opt_plist
    opt_datetime = imap4.IMAP4Server.opt_datetime

    unauth_LOGIN = (do_LOGIN, arg_astring, arg_astring)

    auth_SELECT = (_selectWork, arg_astring, 1, 'SELECT')
    select_SELECT = auth_SELECT

    auth_CREATE = (do_CREATE, arg_astring)
    select_CREATE = auth_CREATE

    auth_EXAMINE = (_selectWork, arg_astring, 0, 'EXAMINE')
    select_EXAMINE = auth_EXAMINE

    auth_DELETE = (do_DELETE, arg_astring)
    select_DELETE = auth_DELETE

    auth_RENAME = (do_RENAME, arg_astring, arg_astring)
    select_RENAME = auth_RENAME

    auth_SUBSCRIBE = (do_SUBSCRIBE, arg_astring)
    select_SUBSCRIBE = auth_SUBSCRIBE

    auth_UNSUBSCRIBE = (do_UNSUBSCRIBE, arg_astring)
    select_UNSUBSCRIBE = auth_UNSUBSCRIBE

    auth_LIST = (_listWork, arg_astring, arg_astring, 0, 'LIST')
    select_LIST = auth_LIST

    auth_LSUB = (_listWork, arg_astring, arg_astring, 1, 'LSUB')
    select_LSUB = auth_LSUB

    auth_STATUS = (do_STATUS, arg_astring, arg_plist)
    select_STATUS = auth_STATUS

    auth_APPEND = (do_APPEND, arg_astring, opt_plist, opt_datetime,
                   arg_literal)
    select_APPEND = auth_APPEND

    select_COPY = (do_COPY, arg_seqset, arg_astring)


    #############################################################
    # END of Twisted imap4 patch to support LITERAL+ extension
    #############################################################



# -*- coding: utf-8 -*-
# test_imap.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Test case for leap.email.imap.server
TestCases taken from twisted tests and modified to make them work
against SoledadBackedAccount.

@authors: Kali Kaneko, <kali@leap.se>
XXX add authors from the original twisted tests.

@license: GPLv3, see included LICENSE file
"""
# XXX review license of the original tests!!!

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

import os
import types


from twisted.mail import imap4
from twisted.internet import defer
from twisted.trial import unittest
from twisted.python import util
from twisted.python import failure

from twisted import cred


# import u1db

from leap.mail.imap.mailbox import SoledadMailbox
from leap.mail.imap.memorystore import MemoryStore
from leap.mail.imap.messages import MessageCollection
from leap.mail.imap.server import LeapIMAPServer
from leap.mail.imap.tests.utils import IMAP4HelperMixin


TEST_USER = "testuser@leap.se"
TEST_PASSWD = "1234"


def strip(f):
    return lambda result, f=f: f()


def sortNest(l):
    l = l[:]
    l.sort()
    for i in range(len(l)):
        if isinstance(l[i], types.ListType):
            l[i] = sortNest(l[i])
        elif isinstance(l[i], types.TupleType):
            l[i] = tuple(sortNest(list(l[i])))
    return l


class TestRealm:

    """
    A minimal auth realm for testing purposes only
    """
    theAccount = None

    def requestAvatar(self, avatarId, mind, *interfaces):
        return imap4.IAccount, self.theAccount, lambda: None


#
# TestCases
#

class MessageCollectionTestCase(IMAP4HelperMixin, unittest.TestCase):

    """
    Tests for the MessageCollection class
    """
    count = 0

    def setUp(self):
        """
        setUp method for each test
        We override mixin method since we are only testing
        MessageCollection interface in this particular TestCase
        """
        super(MessageCollectionTestCase, self).setUp()
        memstore = MemoryStore()
        self.messages = MessageCollection("testmbox%s" % (self.count,),
                                          self._soledad, memstore=memstore)
        MessageCollectionTestCase.count += 1

    def tearDown(self):
        """
        tearDown method for each test
        """
        del self.messages

    def testEmptyMessage(self):
        """
        Test empty message and collection
        """
        em = self.messages._get_empty_doc()
        self.assertEqual(
            em,
            {
                "chash": '',
                "deleted": False,
                "flags": [],
                "mbox": "inbox",
                "seen": False,
                "multi": False,
                "size": 0,
                "type": "flags",
                "uid": 1,
            })
        self.assertEqual(self.messages.count(), 0)

    def testMultipleAdd(self):
        """
        Add multiple messages
        """
        mc = self.messages
        self.assertEqual(self.messages.count(), 0)

        def add_first():
            d = defer.gatherResults([
                mc.add_msg('Stuff 1', subject="test1"),
                mc.add_msg('Stuff 2', subject="test2"),
                mc.add_msg('Stuff 3', subject="test3"),
                mc.add_msg('Stuff 4', subject="test4")])
            return d

        def add_second(result):
            d = defer.gatherResults([
                mc.add_msg('Stuff 5', subject="test5"),
                mc.add_msg('Stuff 6', subject="test6"),
                mc.add_msg('Stuff 7', subject="test7")])
            return d

        def check_second(result):
            return self.assertEqual(mc.count(), 7)

        d1 = add_first()
        d1.addCallback(add_second)
        d1.addCallback(check_second)

    def testRecentCount(self):
        """
        Test the recent count
        """
        mc = self.messages
        countrecent = mc.count_recent
        eq = self.assertEqual

        self.assertEqual(countrecent(), 0)

        d = mc.add_msg('Stuff', subject="test1")
        # For the semantics defined in the RFC, we auto-add the
        # recent flag by default.

        def add2(_):
            return mc.add_msg('Stuff', subject="test2",
                              flags=('\\Deleted',))

        def add3(_):
            return mc.add_msg('Stuff', subject="test3",
                              flags=('\\Recent',))

        def add4(_):
            return mc.add_msg('Stuff', subject="test4",
                              flags=('\\Deleted', '\\Recent'))

        d.addCallback(lambda r: eq(countrecent(), 1))
        d.addCallback(add2)
        d.addCallback(lambda r: eq(countrecent(), 2))
        d.addCallback(add3)
        d.addCallback(lambda r: eq(countrecent(), 3))
        d.addCallback(add4)
        d.addCallback(lambda r: eq(countrecent(), 4))

    def testFilterByMailbox(self):
        """
        Test that queries filter by selected mailbox
        """
        mc = self.messages
        self.assertEqual(self.messages.count(), 0)

        def add_1():
            d1 = mc.add_msg('msg 1', subject="test1")
            d2 = mc.add_msg('msg 2', subject="test2")
            d3 = mc.add_msg('msg 3', subject="test3")
            d = defer.gatherResults([d1, d2, d3])
            return d

        add_1().addCallback(lambda ignored: self.assertEqual(
                            mc.count(), 3))

        # XXX this has to be redone to fit memstore ------------#
        #newmsg = mc._get_empty_doc()
        #newmsg['mailbox'] = "mailbox/foo"
        #mc._soledad.create_doc(newmsg)
        #self.assertEqual(mc.count(), 3)
        #self.assertEqual(
            #len(mc._soledad.get_from_index(mc.TYPE_IDX, "flags")), 4)


class LeapIMAP4ServerTestCase(IMAP4HelperMixin, unittest.TestCase):
    # TODO this currently will use a memory-only store.
    # create a different one for testing soledad sync.

    """
    Tests for the generic behavior of the LeapIMAP4Server
    which, right now, it's just implemented in this test file as
    LeapIMAPServer. We will move the implementation, together with
    authentication bits, to leap.mail.imap.server so it can be instantiated
    from the tac file.

    Right now this TestCase tries to mimmick as close as possible the
    organization from the twisted.mail.imap tests so we can achieve
    a complete implementation. The order in which they appear reflect
    the intended order of implementation.
    """

    #
    # mailboxes operations
    #

    def testCreate(self):
        """
        Test whether we can create mailboxes
        """
        succeed = ('testbox', 'test/box', 'test/', 'test/box/box', 'foobox')
        fail = ('testbox', 'test/box')

        def cb():
            self.result.append(1)

        def eb(failure):
            self.result.append(0)

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def create():
            for name in succeed + fail:
                d = self.client.create(name)
                d.addCallback(strip(cb)).addErrback(eb)
            d.addCallbacks(self._cbStopClient, self._ebGeneral)

        self.result = []
        d1 = self.connected.addCallback(strip(login)).addCallback(
            strip(create))
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        return d.addCallback(self._cbTestCreate, succeed, fail)

    def _cbTestCreate(self, ignored, succeed, fail):
        self.assertEqual(self.result, [1] * len(succeed) + [0] * len(fail))

        mboxes = list(LeapIMAPServer.theAccount.mailboxes)
        answers = ([u'INBOX', u'foobox', 'test', u'test/box',
                    u'test/box/box', 'testbox'])
        self.assertEqual(mboxes, [a for a in answers])

    def testDelete(self):
        """
        Test whether we can delete mailboxes
        """
        LeapIMAPServer.theAccount.addMailbox('delete/me')

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def delete():
            return self.client.delete('delete/me')

        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(delete), self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        d.addCallback(
            lambda _: self.assertEqual(
                LeapIMAPServer.theAccount.mailboxes, ['INBOX']))
        return d

    def testIllegalInboxDelete(self):
        """
        Test what happens if we try to delete the user Inbox.
        We expect that operation to fail.
        """
        self.stashed = None

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def delete():
            return self.client.delete('inbox')

        def stash(result):
            self.stashed = result

        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(delete), self._ebGeneral)
        d1.addBoth(stash)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        d.addCallback(lambda _: self.failUnless(isinstance(self.stashed,
                                                           failure.Failure)))
        return d

    def testNonExistentDelete(self):
        """
        Test what happens if we try to delete a non-existent mailbox.
        We expect an error raised stating 'No such mailbox'
        """
        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def delete():
            return self.client.delete('delete/me')
            self.failure = failure

        def deleteFailed(failure):
            self.failure = failure

        self.failure = None
        d1 = self.connected.addCallback(strip(login))
        d1.addCallback(strip(delete)).addErrback(deleteFailed)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        d.addCallback(lambda _: self.assertTrue(
            str(self.failure.value).startswith('No such mailbox')))
        return d

    def testIllegalDelete(self):
        """
        Try deleting a mailbox with sub-folders, and \NoSelect flag set.
        An exception is expected.
        """
        LeapIMAPServer.theAccount.addMailbox('delete')
        to_delete = LeapIMAPServer.theAccount.getMailbox('delete')
        to_delete.setFlags((r'\Noselect',))
        to_delete.getFlags()
        LeapIMAPServer.theAccount.addMailbox('delete/me')

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def delete():
            return self.client.delete('delete')

        def deleteFailed(failure):
            self.failure = failure

        self.failure = None
        d1 = self.connected.addCallback(strip(login))
        d1.addCallback(strip(delete)).addErrback(deleteFailed)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        expected = ("Hierarchically inferior mailboxes exist "
                    "and \\Noselect is set")
        d.addCallback(lambda _:
                      self.assertTrue(self.failure is not None))
        d.addCallback(lambda _:
                      self.assertEqual(str(self.failure.value), expected))
        return d

    def testRename(self):
        """
        Test whether we can rename a mailbox
        """
        LeapIMAPServer.theAccount.addMailbox('oldmbox')

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def rename():
            return self.client.rename('oldmbox', 'newname')

        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(rename), self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        d.addCallback(lambda _:
                      self.assertEqual(
                          LeapIMAPServer.theAccount.mailboxes,
                          ['INBOX', 'newname']))
        return d

    def testIllegalInboxRename(self):
        """
        Try to rename inbox. We expect it to fail. Then it would be not
        an inbox anymore, would it?
        """
        self.stashed = None

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def rename():
            return self.client.rename('inbox', 'frotz')

        def stash(stuff):
            self.stashed = stuff

        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(rename), self._ebGeneral)
        d1.addBoth(stash)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        d.addCallback(lambda _:
                      self.failUnless(isinstance(
                          self.stashed, failure.Failure)))
        return d

    def testHierarchicalRename(self):
        """
        Try to rename hierarchical mailboxes
        """
        LeapIMAPServer.theAccount.create('oldmbox/m1')
        LeapIMAPServer.theAccount.create('oldmbox/m2')

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def rename():
            return self.client.rename('oldmbox', 'newname')

        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(rename), self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        return d.addCallback(self._cbTestHierarchicalRename)

    def _cbTestHierarchicalRename(self, ignored):
        mboxes = LeapIMAPServer.theAccount.mailboxes
        expected = ['INBOX', 'newname', 'newname/m1', 'newname/m2']
        self.assertEqual(mboxes, [s for s in expected])

    def testSubscribe(self):
        """
        Test whether we can mark a mailbox as subscribed to
        """
        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def subscribe():
            return self.client.subscribe('this/mbox')

        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(subscribe), self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        d.addCallback(lambda _:
                      self.assertEqual(
                          LeapIMAPServer.theAccount.subscriptions,
                          ['this/mbox']))
        return d

    def testUnsubscribe(self):
        """
        Test whether we can unsubscribe from a set of mailboxes
        """
        LeapIMAPServer.theAccount.subscribe('this/mbox')
        LeapIMAPServer.theAccount.subscribe('that/mbox')

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def unsubscribe():
            return self.client.unsubscribe('this/mbox')

        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(unsubscribe), self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        d.addCallback(lambda _:
                      self.assertEqual(
                          LeapIMAPServer.theAccount.subscriptions,
                          ['that/mbox']))
        return d

    def testSelect(self):
        """
        Try to select a mailbox
        """
        self.server.theAccount.addMailbox('TESTMAILBOX-SELECT', creation_ts=42)
        self.selectedArgs = None

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def select():
            def selected(args):
                self.selectedArgs = args
                self._cbStopClient(None)
            d = self.client.select('TESTMAILBOX-SELECT')
            d.addCallback(selected)
            return d

        d1 = self.connected.addCallback(strip(login))
        d1.addCallback(strip(select))
        d1.addErrback(self._ebGeneral)

        d2 = self.loopback()
        return defer.gatherResults([d1, d2]).addCallback(self._cbTestSelect)

    def _cbTestSelect(self, ignored):
        mbox = LeapIMAPServer.theAccount.getMailbox('TESTMAILBOX-SELECT')
        self.assertEqual(self.server.mbox.messages.mbox, mbox.messages.mbox)
        # XXX UIDVALIDITY should be "42" if the creation_ts is passed along
        # to the memory store. However, the current state of the account
        # implementation is incomplete and we're writing to soledad store
        # directly there. We should handle the UIDVALIDITY timestamping
        # mechanism in a separate test suite.

        self.assertEqual(self.selectedArgs, {
            'EXISTS': 0, 'RECENT': 0, 'UIDVALIDITY': 0,
            # 'EXISTS': 0, 'RECENT': 0, 'UIDVALIDITY': 42,
            'FLAGS': ('\\Seen', '\\Answered', '\\Flagged',
                      '\\Deleted', '\\Draft', '\\Recent', 'List'),
            'READ-WRITE': True
        })

    #
    # capabilities
    #

    def testCapability(self):
        caps = {}

        def getCaps():
            def gotCaps(c):
                caps.update(c)
                self.server.transport.loseConnection()
            return self.client.getCapabilities().addCallback(gotCaps)
        d1 = self.connected.addCallback(
            strip(getCaps)).addErrback(self._ebGeneral)
        d = defer.gatherResults([self.loopback(), d1])
        expected = {'IMAP4rev1': None, 'NAMESPACE': None, 'LITERAL+': None,
                    'IDLE': None}

        return d.addCallback(lambda _: self.assertEqual(expected, caps))

    def testCapabilityWithAuth(self):
        caps = {}
        self.server.challengers[
            'CRAM-MD5'] = cred.credentials.CramMD5Credentials

        def getCaps():
            def gotCaps(c):
                caps.update(c)
                self.server.transport.loseConnection()
            return self.client.getCapabilities().addCallback(gotCaps)
        d1 = self.connected.addCallback(
            strip(getCaps)).addErrback(self._ebGeneral)

        d = defer.gatherResults([self.loopback(), d1])

        expCap = {'IMAP4rev1': None, 'NAMESPACE': None,
                  'IDLE': None, 'LITERAL+': None,
                  'AUTH': ['CRAM-MD5']}

        return d.addCallback(lambda _: self.assertEqual(expCap, caps))

    #
    # authentication
    #

    def testLogout(self):
        """
        Test log out
        """
        self.loggedOut = 0

        def logout():
            def setLoggedOut():
                self.loggedOut = 1
            self.client.logout().addCallback(strip(setLoggedOut))
        self.connected.addCallback(strip(logout)).addErrback(self._ebGeneral)
        d = self.loopback()
        return d.addCallback(lambda _: self.assertEqual(self.loggedOut, 1))

    def testNoop(self):
        """
        Test noop command
        """
        self.responses = None

        def noop():
            def setResponses(responses):
                self.responses = responses
                self.server.transport.loseConnection()
            self.client.noop().addCallback(setResponses)
        self.connected.addCallback(strip(noop)).addErrback(self._ebGeneral)
        d = self.loopback()
        return d.addCallback(lambda _: self.assertEqual(self.responses, []))

    def testLogin(self):
        """
        Test login
        """
        def login():
            d = self.client.login(TEST_USER, TEST_PASSWD)
            d.addCallback(self._cbStopClient)
        d1 = self.connected.addCallback(
            strip(login)).addErrback(self._ebGeneral)
        d = defer.gatherResults([d1, self.loopback()])
        return d.addCallback(self._cbTestLogin)

    def _cbTestLogin(self, ignored):
        self.assertEqual(self.server.account, LeapIMAPServer.theAccount)
        self.assertEqual(self.server.state, 'auth')

    def testFailedLogin(self):
        """
        Test bad login
        """
        def login():
            d = self.client.login("bad_user@leap.se", TEST_PASSWD)
            d.addBoth(self._cbStopClient)

        d1 = self.connected.addCallback(
            strip(login)).addErrback(self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        return d.addCallback(self._cbTestFailedLogin)

    def _cbTestFailedLogin(self, ignored):
        self.assertEqual(self.server.state, 'unauth')
        self.assertEqual(self.server.account, None)

    def testLoginRequiringQuoting(self):
        """
        Test login requiring quoting
        """
        self.server._userid = '{test}user@leap.se'
        self.server._password = '{test}password'

        def login():
            d = self.client.login('{test}user@leap.se', '{test}password')
            d.addBoth(self._cbStopClient)

        d1 = self.connected.addCallback(
            strip(login)).addErrback(self._ebGeneral)
        d = defer.gatherResults([self.loopback(), d1])
        return d.addCallback(self._cbTestLoginRequiringQuoting)

    def _cbTestLoginRequiringQuoting(self, ignored):
        self.assertEqual(self.server.account, LeapIMAPServer.theAccount)
        self.assertEqual(self.server.state, 'auth')

    #
    # Inspection
    #

    def testNamespace(self):
        """
        Test retrieving namespace
        """
        self.namespaceArgs = None

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def namespace():
            def gotNamespace(args):
                self.namespaceArgs = args
                self._cbStopClient(None)
            return self.client.namespace().addCallback(gotNamespace)

        d1 = self.connected.addCallback(strip(login))
        d1.addCallback(strip(namespace))
        d1.addErrback(self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        d.addCallback(lambda _: self.assertEqual(self.namespaceArgs,
                                                 [[['', '/']], [], []]))
        return d

    def testExamine(self):
        """
        L{IMAP4Client.examine} issues an I{EXAMINE} command to the server and
        returns a L{Deferred} which fires with a C{dict} with as many of the
        following keys as the server includes in its response: C{'FLAGS'},
        C{'EXISTS'}, C{'RECENT'}, C{'UNSEEN'}, C{'READ-WRITE'}, C{'READ-ONLY'},
        C{'UIDVALIDITY'}, and C{'PERMANENTFLAGS'}.

        Unfortunately the server doesn't generate all of these so it's hard to
        test the client's handling of them here.  See
        L{IMAP4ClientExamineTests} below.

        See U{RFC 3501<http://www.faqs.org/rfcs/rfc3501.html>}, section 6.3.2,
        for details.
        """
        # TODO implement the IMAP4ClientExamineTests testcase.

        self.server.theAccount.addMailbox('test-mailbox-e',
                                          creation_ts=42)
        self.examinedArgs = None

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def examine():
            def examined(args):
                self.examinedArgs = args
                self._cbStopClient(None)
            d = self.client.examine('test-mailbox-e')
            d.addCallback(examined)
            return d

        d1 = self.connected.addCallback(strip(login))
        d1.addCallback(strip(examine))
        d1.addErrback(self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        return d.addCallback(self._cbTestExamine)

    def _cbTestExamine(self, ignored):
        mbox = self.server.theAccount.getMailbox('test-mailbox-e')
        self.assertEqual(self.server.mbox.messages.mbox, mbox.messages.mbox)

        # XXX UIDVALIDITY should be "42" if the creation_ts is passed along
        # to the memory store. However, the current state of the account
        # implementation is incomplete and we're writing to soledad store
        # directly there. We should handle the UIDVALIDITY timestamping
        # mechanism in a separate test suite.
        self.assertEqual(self.examinedArgs, {
            'EXISTS': 0, 'RECENT': 0, 'UIDVALIDITY': 0,
            # 'EXISTS': 0, 'RECENT': 0, 'UIDVALIDITY': 42,
            'FLAGS': ('\\Seen', '\\Answered', '\\Flagged',
                      '\\Deleted', '\\Draft', '\\Recent', 'List'),
            'READ-WRITE': False})

    def _listSetup(self, f):
        LeapIMAPServer.theAccount.addMailbox('root/subthingl',
                                             creation_ts=42)
        LeapIMAPServer.theAccount.addMailbox('root/another-thing',
                                             creation_ts=42)
        LeapIMAPServer.theAccount.addMailbox('non-root/subthing',
                                             creation_ts=42)

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def listed(answers):
            self.listed = answers

        self.listed = None
        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(f), self._ebGeneral)
        d1.addCallbacks(listed, self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        return defer.gatherResults([d1, d2]).addCallback(lambda _: self.listed)

    def testList(self):
        """
        Test List command
        """
        def list():
            return self.client.list('root', '%')
        d = self._listSetup(list)
        d.addCallback(lambda listed: self.assertEqual(
            sortNest(listed),
            sortNest([
                (SoledadMailbox.INIT_FLAGS, "/", "root/subthingl"),
                (SoledadMailbox.INIT_FLAGS, "/", "root/another-thing")
            ])
        ))
        return d

    def testLSub(self):
        """
        Test LSub command
        """
        LeapIMAPServer.theAccount.subscribe('root/subthingl2')

        def lsub():
            return self.client.lsub('root', '%')
        d = self._listSetup(lsub)
        d.addCallback(self.assertEqual,
                      [(SoledadMailbox.INIT_FLAGS, "/", "root/subthingl2")])
        return d

    def testStatus(self):
        """
        Test Status command
        """
        LeapIMAPServer.theAccount.addMailbox('root/subthings')
        # XXX FIXME ---- should populate this a little bit,
        # with unseen etc...

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def status():
            return self.client.status(
                'root/subthings', 'MESSAGES', 'UIDNEXT', 'UNSEEN')

        def statused(result):
            self.statused = result

        self.statused = None
        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(status), self._ebGeneral)
        d1.addCallbacks(statused, self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        d.addCallback(lambda _: self.assertEqual(
            self.statused,
            {'MESSAGES': 0, 'UIDNEXT': '1', 'UNSEEN': 0}
        ))
        return d

    def testFailedStatus(self):
        """
        Test failed status command with a non-existent mailbox
        """
        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def status():
            return self.client.status(
                'root/nonexistent', 'MESSAGES', 'UIDNEXT', 'UNSEEN')

        def statused(result):
            self.statused = result

        def failed(failure):
            self.failure = failure

        self.statused = self.failure = None
        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(status), self._ebGeneral)
        d1.addCallbacks(statused, failed)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        return defer.gatherResults([d1, d2]).addCallback(
            self._cbTestFailedStatus)

    def _cbTestFailedStatus(self, ignored):
        self.assertEqual(
            self.statused, None
        )
        self.assertEqual(
            self.failure.value.args,
            ('Could not open mailbox',)
        )

    #
    # messages
    #

    def testFullAppend(self):
        """
        Test appending a full message to the mailbox
        """
        infile = util.sibpath(__file__, 'rfc822.message')
        message = open(infile)
        LeapIMAPServer.theAccount.addMailbox('root/subthing')

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def append():
            return self.client.append(
                'root/subthing',
                message,
                ('\\SEEN', '\\DELETED'),
                'Tue, 17 Jun 2003 11:22:16 -0600 (MDT)',
            )

        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(append), self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        return d.addCallback(self._cbTestFullAppend, infile)

    def _cbTestFullAppend(self, ignored, infile):
        mb = LeapIMAPServer.theAccount.getMailbox('root/subthing')
        self.assertEqual(1, len(mb.messages))

        msg = mb.messages.get_msg_by_uid(1)
        self.assertEqual(
            set(('\\Recent', '\\SEEN', '\\DELETED')),
            set(msg.getFlags()))

        self.assertEqual(
            'Tue, 17 Jun 2003 11:22:16 -0600 (MDT)',
            msg.getInternalDate())

        parsed = self.parser.parse(open(infile))
        body = parsed.get_payload()
        headers = dict(parsed.items())
        self.assertEqual(
            body,
            msg.getBodyFile().read())
        gotheaders = msg.getHeaders(True)

        self.assertItemsEqual(
            headers, gotheaders)

    def testPartialAppend(self):
        """
        Test partially appending a message to the mailbox
        """
        infile = util.sibpath(__file__, 'rfc822.message')
        LeapIMAPServer.theAccount.addMailbox('PARTIAL/SUBTHING')

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def append():
            message = file(infile)
            return self.client.sendCommand(
                imap4.Command(
                    'APPEND',
                    'PARTIAL/SUBTHING (\\SEEN) "Right now" '
                    '{%d}' % os.path.getsize(infile),
                    (), self.client._IMAP4Client__cbContinueAppend, message
                )
            )
        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(append), self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        return d.addCallback(
            self._cbTestPartialAppend, infile)

    def _cbTestPartialAppend(self, ignored, infile):
        mb = LeapIMAPServer.theAccount.getMailbox('PARTIAL/SUBTHING')
        self.assertEqual(1, len(mb.messages))
        msg = mb.messages.get_msg_by_uid(1)
        self.assertEqual(
            set(('\\SEEN', '\\Recent')),
            set(msg.getFlags())
        )
        parsed = self.parser.parse(open(infile))
        body = parsed.get_payload()
        self.assertEqual(
            body,
            msg.getBodyFile().read())

    def testCheck(self):
        """
        Test check command
        """
        LeapIMAPServer.theAccount.addMailbox('root/subthing')

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def select():
            return self.client.select('root/subthing')

        def check():
            return self.client.check()

        d = self.connected.addCallback(strip(login))
        d.addCallbacks(strip(select), self._ebGeneral)
        d.addCallbacks(strip(check), self._ebGeneral)
        d.addCallbacks(self._cbStopClient, self._ebGeneral)
        return self.loopback()

        # Okay, that was fun

    def testClose(self):
        """
        Test closing the mailbox. We expect to get deleted all messages flagged
        as such.
        """
        name = 'mailbox-close'
        self.server.theAccount.addMailbox(name)

        m = LeapIMAPServer.theAccount.getMailbox(name)

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def select():
            return self.client.select(name)

        def add_messages():
            d1 = m.messages.add_msg(
                'test 1', subject="Message 1",
                flags=('\\Deleted', 'AnotherFlag'))
            d2 = m.messages.add_msg(
                'test 2', subject="Message 2",
                flags=('AnotherFlag',))
            d3 = m.messages.add_msg(
                'test 3', subject="Message 3",
                flags=('\\Deleted',))
            d = defer.gatherResults([d1, d2, d3])
            return d

        def close():
            return self.client.close()

        d = self.connected.addCallback(strip(login))
        d.addCallbacks(strip(select), self._ebGeneral)
        d.addCallbacks(strip(add_messages), self._ebGeneral)
        d.addCallbacks(strip(close), self._ebGeneral)
        d.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        return defer.gatherResults([d, d2]).addCallback(self._cbTestClose, m)

    def _cbTestClose(self, ignored, m):
        self.assertEqual(len(m.messages), 1)
        msg = m.messages.get_msg_by_uid(2)
        self.assertTrue(msg is not None)

        self.assertEqual(
            dict(msg.hdoc.content)['subject'],
            'Message 2')
        self.failUnless(m.closed)

    def testExpunge(self):
        """
        Test expunge command
        """
        name = 'mailbox-expunge'
        self.server.theAccount.addMailbox(name)
        m = LeapIMAPServer.theAccount.getMailbox(name)

        def login():
            return self.client.login(TEST_USER, TEST_PASSWD)

        def select():
            return self.client.select('mailbox-expunge')

        def add_messages():
            d1 = m.messages.add_msg(
                'test 1', subject="Message 1",
                flags=('\\Deleted', 'AnotherFlag'))
            d2 = m.messages.add_msg(
                'test 2', subject="Message 2",
                flags=('AnotherFlag',))
            d3 = m.messages.add_msg(
                'test 3', subject="Message 3",
                flags=('\\Deleted',))
            d = defer.gatherResults([d1, d2, d3])
            return d

        def expunge():
            return self.client.expunge()

        def expunged(results):
            self.failIf(self.server.mbox is None)
            self.results = results

        self.results = None
        d1 = self.connected.addCallback(strip(login))
        d1.addCallbacks(strip(select), self._ebGeneral)
        d1.addCallbacks(strip(add_messages), self._ebGeneral)
        d1.addCallbacks(strip(expunge), self._ebGeneral)
        d1.addCallbacks(expunged, self._ebGeneral)
        d1.addCallbacks(self._cbStopClient, self._ebGeneral)
        d2 = self.loopback()
        d = defer.gatherResults([d1, d2])
        return d.addCallback(self._cbTestExpunge, m)

    def _cbTestExpunge(self, ignored, m):
        # we only left 1 mssage with no deleted flag
        self.assertEqual(len(m.messages), 1)
        msg = m.messages.get_msg_by_uid(2)

        msg = list(m.messages)[0]
        self.assertTrue(msg is not None)

        self.assertEqual(
            msg.hdoc.content['subject'],
            'Message 2')

        # the uids of the deleted messages
        self.assertItemsEqual(self.results, [1, 3])


class AccountTestCase(IMAP4HelperMixin, unittest.TestCase):
    """
    Test the Account.
    """
    def _create_empty_mailbox(self):
        LeapIMAPServer.theAccount.addMailbox('')

    def _create_one_mailbox(self):
        LeapIMAPServer.theAccount.addMailbox('one')

    def test_illegalMailboxCreate(self):
        self.assertRaises(AssertionError, self._create_empty_mailbox)


class IMAP4ServerSearchTestCase(IMAP4HelperMixin, unittest.TestCase):
    """
    Tests for the behavior of the search_* functions in L{imap5.IMAP4Server}.
    """
    # XXX coming soon to your screens!
    pass



# -*- coding: utf-8 -*-
# soledadstore.py
# Copyright (C) 2014 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
A MessageStore that writes to Soledad.
"""
import logging
import threading

from collections import defaultdict
from itertools import chain

from u1db import errors as u1db_errors
from twisted.python import log
from zope.interface import implements

from leap.common.check import leap_assert_type, leap_assert
from leap.mail.decorators import deferred_to_thread
from leap.mail.imap.messageparts import MessagePartType
from leap.mail.imap.messageparts import MessageWrapper
from leap.mail.imap.messageparts import RecentFlagsDoc
from leap.mail.imap.fields import fields
from leap.mail.imap.interfaces import IMessageStore
from leap.mail.messageflow import IMessageConsumer
from leap.mail.utils import first, empty, accumulator_queue

logger = logging.getLogger(__name__)


# TODO
# [ ] Implement a retry queue?
# [ ] Consider journaling of operations.


class ContentDedup(object):
    """
    Message deduplication.

    We do a query for the content hashes before writing to our beloved
    sqlcipher backend of Soledad. This means, by now, that:

    1. We will not store the same body/attachment twice, only the hash of it.
    2. We will not store the same message header twice, only the hash of it.

    The first case is useful if you are always receiving the same old memes
    from unwary friends that still have not discovered that 4chan is the
    generator of the internet. The second will save your day if you have
    initiated session with the same account in two different machines. I also
    wonder why would you do that, but let's respect each other choices, like
    with the religious celebrations, and assume that one day we'll be able
    to run Bitmask in completely free phones. Yes, I mean that, the whole GSM
    Stack.
    """
    # TODO refactor using unique_query

    def _header_does_exist(self, doc):
        """
        Check whether we already have a header document for this
        content hash in our database.

        :param doc: tentative header for document
        :type doc: dict
        :returns: True if it exists, False otherwise.
        """
        if not doc:
            return False
        chash = doc[fields.CONTENT_HASH_KEY]
        header_docs = self._soledad.get_from_index(
            fields.TYPE_C_HASH_IDX,
            fields.TYPE_HEADERS_VAL, str(chash))
        if not header_docs:
            return False

        # FIXME enable only to debug this problem.
        #if len(header_docs) != 1:
            #logger.warning("Found more than one copy of chash %s!"
                           #% (chash,))

        #logger.debug("Found header doc with that hash! Skipping save!")
        return True

    def _content_does_exist(self, doc):
        """
        Check whether we already have a content document for a payload
        with this hash in our database.

        :param doc: tentative content for document
        :type doc: dict
        :returns: True if it exists, False otherwise.
        """
        if not doc:
            return False
        phash = doc[fields.PAYLOAD_HASH_KEY]
        attach_docs = self._soledad.get_from_index(
            fields.TYPE_P_HASH_IDX,
            fields.TYPE_CONTENT_VAL, str(phash))
        if not attach_docs:
            return False

        # FIXME enable only to debug this problem
        #if len(attach_docs) != 1:
            #logger.warning("Found more than one copy of phash %s!"
                           #% (phash,))
        #logger.debug("Found attachment doc with that hash! Skipping save!")
        return True


class MsgWriteError(Exception):
    """
    Raised if any exception is found while saving message parts.
    """
    pass


"""
A lock per document.
"""
# TODO should bound the space of this!!!
# http://stackoverflow.com/a/2437645/1157664
# Setting this to twice the number of threads in the threadpool
# should be safe.
put_locks = defaultdict(lambda: threading.Lock())
mbox_doc_locks = defaultdict(lambda: threading.Lock())


class SoledadStore(ContentDedup):
    """
    This will create docs in the local Soledad database.
    """
    _remove_lock = threading.Lock()

    implements(IMessageConsumer, IMessageStore)

    def __init__(self, soledad):
        """
        Initialize the permanent store that writes to Soledad database.

        :param soledad: the soledad instance
        :type soledad: Soledad
        """
        from twisted.internet import reactor
        self.reactor = reactor

        self._soledad = soledad

        self._CREATE_DOC_FUN = self._soledad.create_doc
        self._PUT_DOC_FUN = self._soledad.put_doc
        self._GET_DOC_FUN = self._soledad.get_doc

        # we instantiate an accumulator to batch the notifications
        self.docs_notify_queue = accumulator_queue(
            lambda item: reactor.callFromThread(self._unset_new_dirty, item),
            20)

    # IMessageStore

    # -------------------------------------------------------------------
    # We are not yet using this interface, but it would make sense
    # to implement it.

    def create_message(self, mbox, uid, message):
        """
        Create the passed message into this SoledadStore.

        :param mbox: the mbox this message belongs.
        :type mbox: str or unicode
        :param uid: the UID that identifies this message in this mailbox.
        :type uid: int
        :param message: a IMessageContainer implementor.
        """
        raise NotImplementedError()

    def put_message(self, mbox, uid, message):
        """
        Put the passed existing message into this SoledadStore.

        :param mbox: the mbox this message belongs.
        :type mbox: str or unicode
        :param uid: the UID that identifies this message in this mailbox.
        :type uid: int
        :param message: a IMessageContainer implementor.
        """
        raise NotImplementedError()

    def remove_message(self, mbox, uid):
        """
        Remove the given message from this SoledadStore.

        :param mbox: the mbox this message belongs.
        :type mbox: str or unicode
        :param uid: the UID that identifies this message in this mailbox.
        :type uid: int
        """
        raise NotImplementedError()

    def get_message(self, mbox, uid):
        """
        Get a IMessageContainer for the given mbox and uid combination.

        :param mbox: the mbox this message belongs.
        :type mbox: str or unicode
        :param uid: the UID that identifies this message in this mailbox.
        :type uid: int
        """
        raise NotImplementedError()

    # IMessageConsumer

    # TODO should handle the delete case
    # TODO should handle errors better
    # TODO could generalize this method into a generic consumer
    # and only implement `process` here

    def consume(self, queue):
        """
        Creates a new document in soledad db.

        :param queue: a tuple of queues to get item from, with content of the
                      document to be inserted.
        :type queue: tuple of Queues
        """
        new, dirty = queue
        while not new.empty():
            doc_wrapper = new.get()
            self.reactor.callInThread(self._consume_doc, doc_wrapper,
                                      self.docs_notify_queue)
        while not dirty.empty():
            doc_wrapper = dirty.get()
            self.reactor.callInThread(self._consume_doc, doc_wrapper,
                                      self.docs_notify_queue)

        # Queue empty, flush the notifications queue.
        self.docs_notify_queue(None, flush=True)

    def _unset_new_dirty(self, doc_wrapper):
        """
        Unset the `new` and `dirty` flags for this document wrapper in the
        memory store.

        :param doc_wrapper: a MessageWrapper instance
        :type doc_wrapper: MessageWrapper
        """
        if isinstance(doc_wrapper, MessageWrapper):
            # XXX still needed for debug quite often
            #logger.info("unsetting new flag!")
            doc_wrapper.new = False
            doc_wrapper.dirty = False

    @deferred_to_thread
    def _consume_doc(self, doc_wrapper, notify_queue):
        """
        Consume each document wrapper in a separate thread.
        We pass an instance of an accumulator that handles the notifications
        to the memorystore when the write has been done.

        :param doc_wrapper: a MessageWrapper or RecentFlagsDoc instance
        :type doc_wrapper: MessageWrapper or RecentFlagsDoc
        :param notify_queue: a callable that handles the writeback
                             notifications to the memstore.
        :type notify_queue: callable
        """
        def queueNotifyBack(failed, doc_wrapper):
            if failed:
                log.msg("There was an error writing the mesage...")
            else:
                notify_queue(doc_wrapper)

        def doSoledadCalls(items):
            # we prime the generator, that should return the
            # message or flags wrapper item in the first place.
            try:
                doc_wrapper = items.next()
            except StopIteration:
                pass
            else:
                failed = self._soledad_write_document_parts(items)
                queueNotifyBack(failed, doc_wrapper)

        doSoledadCalls(self._iter_wrapper_subparts(doc_wrapper))

    #
    # SoledadStore specific methods.
    #

    def _soledad_write_document_parts(self, items):
        """
        Write the document parts to soledad in a separate thread.

        :param items: the iterator through the different document wrappers
                      payloads.
        :type items: iterator
        :return: whether the write was successful or not
        :rtype: bool
        """
        failed = False
        for item, call in items:
            if empty(item):
                continue
            try:
                self._try_call(call, item)
            except Exception as exc:
                logger.debug("ITEM WAS: %s" % repr(item))
                if hasattr(item, 'content'):
                    logger.debug("ITEM CONTENT WAS: %s" %
                                 repr(item.content))
                logger.exception(exc)
                failed = True
                continue
        return failed

    def _iter_wrapper_subparts(self, doc_wrapper):
        """
        Return an iterator that will yield the doc_wrapper in the first place,
        followed by the subparts item and the proper call type for every
        item in the queue, if any.

        :param doc_wrapper: a MessageWrapper or RecentFlagsDoc instance
        :type doc_wrapper: MessageWrapper or RecentFlagsDoc
        """
        if isinstance(doc_wrapper, MessageWrapper):
            return chain((doc_wrapper,),
                         self._get_calls_for_msg_parts(doc_wrapper))
        elif isinstance(doc_wrapper, RecentFlagsDoc):
            return chain((doc_wrapper,),
                         self._get_calls_for_rflags_doc(doc_wrapper))
        else:
            logger.warning("CANNOT PROCESS ITEM!")
            return (i for i in [])

    def _try_call(self, call, item):
        """
        Try to invoke a given call with item as a parameter.

        :param call: the function to call
        :type call: callable
        :param item: the payload to pass to the call as argument
        :type item: object
        """
        if call is None:
            return

        if call == self._PUT_DOC_FUN:
            doc_id = item.doc_id
            if doc_id is None:
                logger.warning("BUG! Dirty doc but has no doc_id!")
                return
            with put_locks[doc_id]:
                doc = self._GET_DOC_FUN(doc_id)

                if doc is None:
                    logger.warning("BUG! Dirty doc but could not "
                                   "find document %s" % (doc_id,))
                    return

                doc.content = dict(item.content)

                item = doc
                try:
                    call(item)
                except u1db_errors.RevisionConflict as exc:
                    logger.exception("Error: %r" % (exc,))
                    raise exc
                except Exception as exc:
                    logger.exception("Error: %r" % (exc,))
                    raise exc

        else:
            try:
                call(item)
            except u1db_errors.RevisionConflict as exc:
                logger.exception("Error: %r" % (exc,))
                raise exc
            except Exception as exc:
                logger.exception("Error: %r" % (exc,))
                raise exc

    def _get_calls_for_msg_parts(self, msg_wrapper):
        """
        Generator that return the proper call type for a given item.

        :param msg_wrapper: A MessageWrapper
        :type msg_wrapper: IMessageContainer
        :return: a generator of tuples with recent-flags doc payload
                 and callable
        :rtype: generator
        """
        call = None

        if msg_wrapper.new:
            call = self._CREATE_DOC_FUN

            # item is expected to be a MessagePartDoc
            for item in msg_wrapper.walk():
                if item.part == MessagePartType.fdoc:
                    yield dict(item.content), call

                elif item.part == MessagePartType.hdoc:
                    if not self._header_does_exist(item.content):
                        yield dict(item.content), call

                elif item.part == MessagePartType.cdoc:
                    if not self._content_does_exist(item.content):
                        yield dict(item.content), call

        # For now, the only thing that will be dirty is
        # the flags doc.

        elif msg_wrapper.dirty:
            call = self._PUT_DOC_FUN
            # item is expected to be a MessagePartDoc
            for item in msg_wrapper.walk():
                # XXX FIXME Give error if dirty and not doc_id !!!
                doc_id = item.doc_id  # defend!
                if not doc_id:
                    logger.warning("Dirty item but no doc_id!")
                    continue

                if item.part == MessagePartType.fdoc:
                    #logger.debug("PUT dirty fdoc")
                    yield item, call

                # XXX also for linkage-doc !!!
        else:
            logger.error("Cannot delete documents yet from the queue...!")

    def _get_calls_for_rflags_doc(self, rflags_wrapper):
        """
        We always put these documents.

        :param rflags_wrapper: A wrapper around recent flags doc.
        :type rflags_wrapper: RecentFlagsWrapper
        :return: a tuple with recent-flags doc payload and callable
        :rtype: tuple
        """
        call = self._PUT_DOC_FUN

        payload = rflags_wrapper.content
        if payload:
            logger.debug("Saving RFLAGS to Soledad...")
            yield rflags_wrapper, call

    # Mbox documents and attributes

    def get_mbox_document(self, mbox):
        """
        Return mailbox document.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :return: A SoledadDocument containing this mailbox, or None if
                 the query failed.
        :rtype: SoledadDocument or None.
        """
        with mbox_doc_locks[mbox]:
            return self._get_mbox_document(mbox)

    def _get_mbox_document(self, mbox):
        """
        Helper for returning the mailbox document.
        """
        try:
            query = self._soledad.get_from_index(
                fields.TYPE_MBOX_IDX,
                fields.TYPE_MBOX_VAL, mbox)
            if query:
                return query.pop()
            else:
                logger.error("Could not find mbox document for %r" %
                            (mbox,))
        except Exception as exc:
            logger.exception("Unhandled error %r" % exc)

    def get_mbox_closed(self, mbox):
        """
        Return the closed attribute for a given mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :rtype: bool
        """
        mbox_doc = self.get_mbox_document()
        return mbox_doc.content.get(fields.CLOSED_KEY, False)

    def set_mbox_closed(self, mbox, closed):
        """
        Set the closed attribute for a given mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param closed:  the value to be set
        :type closed: bool
        """
        leap_assert(isinstance(closed, bool), "closed needs to be boolean")
        with mbox_doc_locks[mbox]:
            mbox_doc = self._get_mbox_document(mbox)
            if mbox_doc is None:
                logger.error(
                    "Could not find mbox document for %r" % (mbox,))
                return
            mbox_doc.content[fields.CLOSED_KEY] = closed
            self._soledad.put_doc(mbox_doc)

    def write_last_uid(self, mbox, value):
        """
        Write the `last_uid` integer to the proper mailbox document
        in Soledad.
        This is called from the deferred triggered by
        memorystore.increment_last_soledad_uid, which is expected to
        run in a separate thread.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param value: the value to set
        :type value: int
        """
        leap_assert_type(value, int)
        key = fields.LAST_UID_KEY

        # XXX use accumulator to reduce number of hits
        with mbox_doc_locks[mbox]:
            mbox_doc = self._get_mbox_document(mbox)
            old_val = mbox_doc.content[key]
            if value > old_val:
                mbox_doc.content[key] = value
                try:
                    self._soledad.put_doc(mbox_doc)
                except Exception as exc:
                    logger.error("Error while setting last_uid for %r"
                                 % (mbox,))
                    logger.exception(exc)

    def get_flags_doc(self, mbox, uid):
        """
        Return the SoledadDocument for the given mbox and uid.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :param uid: the UID for the message
        :type uid: int
        :rtype: SoledadDocument or None
        """
        result = None
        try:
            flag_docs = self._soledad.get_from_index(
                fields.TYPE_MBOX_UID_IDX,
                fields.TYPE_FLAGS_VAL, mbox, str(uid))
            if len(flag_docs) != 1:
                logger.warning("More than one flag doc for %r:%s" %
                               (mbox, uid))
            result = first(flag_docs)
        except Exception as exc:
            # ugh! Something's broken down there!
            logger.warning("ERROR while getting flags for UID: %s" % uid)
            logger.exception(exc)
        finally:
            return result

    def get_headers_doc(self, chash):
        """
        Return the document that keeps the headers for a message
        indexed by its content-hash.

        :param chash: the content-hash to retrieve the document from.
        :type chash: str or unicode
        :rtype: SoledadDocument or None
        """
        head_docs = self._soledad.get_from_index(
            fields.TYPE_C_HASH_IDX,
            fields.TYPE_HEADERS_VAL, str(chash))
        return first(head_docs)

    # deleted messages

    def deleted_iter(self, mbox):
        """
        Get an iterator for the the doc_id for SoledadDocuments for messages
        with \\Deleted flag for a given mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        :return: iterator through deleted message docs
        :rtype: iterable
        """
        return [doc.doc_id for doc in self._soledad.get_from_index(
                fields.TYPE_MBOX_DEL_IDX,
                fields.TYPE_FLAGS_VAL, mbox, '1')]

    def remove_all_deleted(self, mbox):
        """
        Remove from Soledad all messages flagged as deleted for a given
        mailbox.

        :param mbox: the mailbox
        :type mbox: str or unicode
        """
        deleted = []
        for doc_id in self.deleted_iter(mbox):
            with self._remove_lock:
                doc = self._soledad.get_doc(doc_id)
                if doc is not None:
                    self._soledad.delete_doc(doc)
                    try:
                        deleted.append(doc.content[fields.UID_KEY])
                    except TypeError:
                        # empty content
                        pass
        return deleted



# -*- coding: utf-8 -*-
# fetch.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Incoming mail fetcher.
"""
import copy
import logging
import threading
import time
import sys
import traceback
import warnings

from email.parser import Parser
from email.generator import Generator
from email.utils import parseaddr
from StringIO import StringIO

from twisted.python import log
from twisted.internet import defer
from twisted.internet.task import LoopingCall
from twisted.internet.task import deferLater
from u1db import errors as u1db_errors
from zope.proxy import sameProxiedObjects

from leap.common import events as leap_events
from leap.common.check import leap_assert, leap_assert_type
from leap.common.events.events_pb2 import IMAP_FETCHED_INCOMING
from leap.common.events.events_pb2 import IMAP_MSG_PROCESSING
from leap.common.events.events_pb2 import IMAP_MSG_DECRYPTED
from leap.common.events.events_pb2 import IMAP_MSG_SAVED_LOCALLY
from leap.common.events.events_pb2 import IMAP_MSG_DELETED_INCOMING
from leap.common.events.events_pb2 import IMAP_UNREAD_MAIL
from leap.common.events.events_pb2 import SOLEDAD_INVALID_AUTH_TOKEN
from leap.common.mail import get_email_charset
from leap.keymanager import errors as keymanager_errors
from leap.keymanager.openpgp import OpenPGPKey
from leap.mail.decorators import deferred_to_thread
from leap.mail.imap.fields import fields
from leap.mail.utils import json_loads, empty, first
from leap.soledad.client import Soledad
from leap.soledad.common.crypto import ENC_SCHEME_KEY, ENC_JSON_KEY
from leap.soledad.common.errors import InvalidAuthTokenError


logger = logging.getLogger(__name__)

MULTIPART_ENCRYPTED = "multipart/encrypted"
MULTIPART_SIGNED = "multipart/signed"
PGP_BEGIN = "-----BEGIN PGP MESSAGE-----"
PGP_END = "-----END PGP MESSAGE-----"


class MalformedMessage(Exception):
    """
    Raised when a given message is not well formed.
    """
    pass


class LeapIncomingMail(object):
    """
    Fetches and process mail from the incoming pool.

    This object has public methods start_loop and stop that will
    actually initiate a LoopingCall with check_period recurrency.
    The LoopingCall itself will invoke the fetch method each time
    that the check_period expires.

    This loop will sync the soledad db with the remote server and
    process all the documents found tagged as incoming mail.
    """

    RECENT_FLAG = "\\Recent"
    CONTENT_KEY = "content"

    LEAP_SIGNATURE_HEADER = 'X-Leap-Signature'
    """
    Header added to messages when they are decrypted by the IMAP fetcher,
    which states the validity of an eventual signature that might be included
    in the encrypted blob.
    """
    LEAP_SIGNATURE_VALID = 'valid'
    LEAP_SIGNATURE_INVALID = 'invalid'
    LEAP_SIGNATURE_COULD_NOT_VERIFY = 'could not verify'

    fetching_lock = threading.Lock()

    def __init__(self, keymanager, soledad, imap_account,
                 check_period, userid):

        """
        Initialize LeapIncomingMail..

        :param keymanager: a keymanager instance
        :type keymanager: keymanager.KeyManager

        :param soledad: a soledad instance
        :type soledad: Soledad

        :param imap_account: the account to fetch periodically
        :type imap_account: SoledadBackedAccount

        :param check_period: the period to fetch new mail, in seconds.
        :type check_period: int
        """

        leap_assert(keymanager, "need a keymanager to initialize")
        leap_assert_type(soledad, Soledad)
        leap_assert(check_period, "need a period to check incoming mail")
        leap_assert_type(check_period, int)
        leap_assert(userid, "need a userid to initialize")

        self._keymanager = keymanager
        self._soledad = soledad
        self.imapAccount = imap_account
        self._inbox = self.imapAccount.getMailbox('inbox')
        self._userid = userid

        self._loop = None
        self._check_period = check_period

        # initialize a mail parser only once
        self._parser = Parser()

    @property
    def _pkey(self):
        if sameProxiedObjects(self._keymanager, None):
            logger.warning('tried to get key, but null keymanager found')
            return None
        return self._keymanager.get_key(self._userid, OpenPGPKey, private=True)

    #
    # Public API: fetch, start_loop, stop.
    #

    def fetch(self):
        """
        Fetch incoming mail, to be called periodically.

        Calls a deferred that will execute the fetch callback
        in a separate thread
        """
        def syncSoledadCallback(result):
            # FIXME this needs a matching change in mx!!!
            # --> need to add ERROR_DECRYPTING_KEY = False
            # as default.
            try:
                doclist = self._soledad.get_from_index(
                    fields.JUST_MAIL_IDX, "*", "0")
            except u1db_errors.InvalidGlobbing:
                # It looks like we are a dealing with an outdated
                # mx. Fallback to the version of the index
                warnings.warn("JUST_MAIL_COMPAT_IDX will be deprecated!",
                              DeprecationWarning)
                doclist = self._soledad.get_from_index(
                    fields.JUST_MAIL_COMPAT_IDX, "*")
            self._process_doclist(doclist)

        logger.debug("fetching mail for: %s %s" % (
            self._soledad.uuid, self._userid))
        if not self.fetching_lock.locked():
            d1 = self._sync_soledad()
            d = defer.gatherResults([d1], consumeErrors=True)
            d.addCallbacks(syncSoledadCallback, self._errback)
            d.addCallbacks(self._signal_fetch_to_ui, self._errback)
            return d
        else:
            logger.debug("Already fetching mail.")

    def start_loop(self):
        """
        Starts a loop to fetch mail.
        """
        if self._loop is None:
            self._loop = LoopingCall(self.fetch)
            self._loop.start(self._check_period)
        else:
            logger.warning("Tried to start an already running fetching loop.")

    def stop(self):
        # XXX change the name to stop_loop, for consistency.
        """
        Stops the loop that fetches mail.
        """
        if self._loop and self._loop.running is True:
            self._loop.stop()
            self._loop = None

    #
    # Private methods.
    #

    # synchronize incoming mail

    def _errback(self, failure):
        logger.exception(failure.value)
        traceback.print_tb(*sys.exc_info())

    @deferred_to_thread
    def _sync_soledad(self):
        """
        Synchronize with remote soledad.

        :returns: a list of LeapDocuments, or None.
        :rtype: iterable or None
        """
        with self.fetching_lock:
            try:
                log.msg('FETCH: syncing soledad...')
                self._soledad.sync()
                log.msg('FETCH soledad SYNCED.')
            except InvalidAuthTokenError:
                # if the token is invalid, send an event so the GUI can
                # disable mail and show an error message.
                leap_events.signal(SOLEDAD_INVALID_AUTH_TOKEN)

    def _signal_fetch_to_ui(self, doclist):
        """
        Send leap events to ui.

        :param doclist: iterable with msg documents.
        :type doclist: iterable.
        :returns: doclist
        :rtype: iterable
        """
        doclist = first(doclist)  # gatherResults pass us a list
        if doclist:
            fetched_ts = time.mktime(time.gmtime())
            num_mails = len(doclist) if doclist is not None else 0
            if num_mails != 0:
                log.msg("there are %s mails" % (num_mails,))
            leap_events.signal(
                IMAP_FETCHED_INCOMING, str(num_mails), str(fetched_ts))
            return doclist

    def _signal_unread_to_ui(self, *args):
        """
        Sends unread event to ui.
        """
        leap_events.signal(
            IMAP_UNREAD_MAIL, str(self._inbox.getUnseenCount()))

    # process incoming mail.

    def _process_doclist(self, doclist):
        """
        Iterates through the doclist, checks if each doc
        looks like a message, and yields a deferred that will decrypt and
        process the message.

        :param doclist: iterable with msg documents.
        :type doclist: iterable.
        :returns: a list of deferreds for individual messages.
        """
        log.msg('processing doclist')
        if not doclist:
            logger.debug("no docs found")
            return
        num_mails = len(doclist)

        for index, doc in enumerate(doclist):
            logger.debug("processing doc %d of %d" % (index + 1, num_mails))
            leap_events.signal(
                IMAP_MSG_PROCESSING, str(index), str(num_mails))

            keys = doc.content.keys()

            # TODO Compatibility check with the index in pre-0.6 mx
            # that does not write the ERROR_DECRYPTING_KEY
            # This should be removed in 0.7

            has_errors = doc.content.get(fields.ERROR_DECRYPTING_KEY, None)
            if has_errors is None:
                warnings.warn("JUST_MAIL_COMPAT_IDX will be deprecated!",
                              DeprecationWarning)
            if has_errors:
                logger.debug("skipping msg with decrypting errors...")

            if self._is_msg(keys) and not has_errors:
                # Evaluating to bool of has_errors is intentional here.
                # We don't mind at this point if it's None or False.

                # Ok, this looks like a legit msg, and with no errors.
                # Let's process it!

                d1 = self._decrypt_doc(doc)
                d = defer.gatherResults([d1], consumeErrors=True)
                d.addCallbacks(self._add_message_locally, self._errback)

    #
    # operations on individual messages
    #

    @deferred_to_thread
    def _decrypt_doc(self, doc):
        """
        Decrypt the contents of a document.

        :param doc: A document containing an encrypted message.
        :type doc: SoledadDocument

        :return: A tuple containing the document and the decrypted message.
        :rtype: (SoledadDocument, str)
        """
        log.msg('decrypting msg')
        success = False

        try:
            decrdata = self._keymanager.decrypt(
                doc.content[ENC_JSON_KEY],
                self._pkey)
            success = True
        except Exception as exc:
            # XXX move this to errback !!!
            logger.error("Error while decrypting msg: %r" % (exc,))
            decrdata = ""
        leap_events.signal(IMAP_MSG_DECRYPTED, "1" if success else "0")

        data = self._process_decrypted_doc((doc, decrdata))
        return (doc, data)

    def _process_decrypted_doc(self, msgtuple):
        """
        Process a document containing a succesfully decrypted message.

        :param msgtuple: a tuple consisting of a SoledadDocument
                         instance containing the incoming message
                         and data, the json-encoded, decrypted content of the
                         incoming message
        :type msgtuple: (SoledadDocument, str)
        :return: a SoledadDocument and the processed data.
        :rtype: (doc, data)
        """
        log.msg('processing decrypted doc')
        doc, data = msgtuple

        from twisted.internet import reactor

        # XXX turn this into an errBack for each one of
        # the deferreds that would process an individual document
        try:
            msg = json_loads(data)
        except UnicodeError as exc:
            logger.error("Error while decrypting %s" % (doc.doc_id,))
            logger.exception(exc)

            # we flag the message as "with decrypting errors",
            # to avoid further decryption attempts during sync
            # cycles until we're prepared to deal with that.
            # What is the same, when Ivan deals with it...
            # A new decrypting attempt event could be triggered by a
            # future a library upgrade, or a cli flag to the client,
            # we just `defer` that for now... :)
            doc.content[fields.ERROR_DECRYPTING_KEY] = True
            deferLater(reactor, 0, self._update_incoming_message, doc)

            # FIXME this is just a dirty hack to delay the proper
            # deferred organization here...
            # and remember, boys, do not do this at home.
            return []

        if not isinstance(msg, dict):
            defer.returnValue(False)
        if not msg.get(fields.INCOMING_KEY, False):
            defer.returnValue(False)

        # ok, this is an incoming message
        rawmsg = msg.get(self.CONTENT_KEY, None)
        if not rawmsg:
            return False
        return self._maybe_decrypt_msg(rawmsg)

    @deferred_to_thread
    def _update_incoming_message(self, doc):
        """
        Do a put for a soledad document. This probably has been called only
        in the case that we've needed to update the ERROR_DECRYPTING_KEY
        flag in an incoming message, to get it out of the decrypting queue.

        :param doc: the SoledadDocument to update
        :type doc: SoledadDocument
        """
        log.msg("Updating SoledadDoc %s" % (doc.doc_id))
        self._soledad.put_doc(doc)

    @deferred_to_thread
    def _delete_incoming_message(self, doc):
        """
        Delete document.

        :param doc: the SoledadDocument to delete
        :type doc: SoledadDocument
        """
        log.msg("Deleting Incoming message: %s" % (doc.doc_id,))
        self._soledad.delete_doc(doc)

    def _maybe_decrypt_msg(self, data):
        """
        Tries to decrypt a gpg message if data looks like one.

        :param data: the text to be decrypted.
        :type data: str
        :return: data, possibly descrypted.
        :rtype: str
        """
        leap_assert_type(data, str)
        log.msg('maybe decrypting doc')

        # parse the original message
        encoding = get_email_charset(data)
        msg = self._parser.parsestr(data)

        # try to obtain sender public key
        senderPubkey = None
        fromHeader = msg.get('from', None)
        if (fromHeader is not None
            and (msg.get_content_type() == MULTIPART_ENCRYPTED
                 or msg.get_content_type() == MULTIPART_SIGNED)):
            _, senderAddress = parseaddr(fromHeader)
            try:
                senderPubkey = self._keymanager.get_key_from_cache(
                    senderAddress, OpenPGPKey)
            except keymanager_errors.KeyNotFound:
                pass

        valid_sig = False  # we will add a header saying if sig is valid
        decrypt_multi = self._decrypt_multipart_encrypted_msg
        decrypt_inline = self._maybe_decrypt_inline_encrypted_msg

        if msg.get_content_type() == MULTIPART_ENCRYPTED:
            decrmsg, valid_sig = decrypt_multi(
                msg, encoding, senderPubkey)
        else:
            decrmsg, valid_sig = decrypt_inline(
                msg, encoding, senderPubkey)

        # add x-leap-signature header
        if senderPubkey is None:
            decrmsg.add_header(
                self.LEAP_SIGNATURE_HEADER,
                self.LEAP_SIGNATURE_COULD_NOT_VERIFY)
        else:
            decrmsg.add_header(
                self.LEAP_SIGNATURE_HEADER,
                self.LEAP_SIGNATURE_VALID if valid_sig else
                self.LEAP_SIGNATURE_INVALID,
                pubkey=senderPubkey.key_id)

        return decrmsg.as_string()

    def _decrypt_multipart_encrypted_msg(self, msg, encoding, senderPubkey):
        """
        Decrypt a message with content-type 'multipart/encrypted'.

        :param msg: The original encrypted message.
        :type msg: Message
        :param encoding: The encoding of the email message.
        :type encoding: str
        :param senderPubkey: The key of the sender of the message.
        :type senderPubkey: OpenPGPKey

        :return: A unitary tuple containing a decrypted message.
        :rtype: (Message)
        """
        log.msg('decrypting multipart encrypted msg')
        msg = copy.deepcopy(msg)
        self._msg_multipart_sanity_check(msg)

        # parse message and get encrypted content
        pgpencmsg = msg.get_payload()[1]
        encdata = pgpencmsg.get_payload()

        # decrypt or fail gracefully
        try:
            decrdata, valid_sig = self._decrypt_and_verify_data(
                encdata, senderPubkey)
        except keymanager_errors.DecryptError as e:
            logger.warning('Failed to decrypt encrypted message (%s). '
                           'Storing message without modifications.' % str(e))
            # Bailing out!
            return (msg, False)

        decrmsg = self._parser.parsestr(decrdata)
        # remove original message's multipart/encrypted content-type
        del(msg['content-type'])

        # replace headers back in original message
        for hkey, hval in decrmsg.items():
            try:
                # this will raise KeyError if header is not present
                msg.replace_header(hkey, hval)
            except KeyError:
                msg[hkey] = hval

        # all ok, replace payload by unencrypted payload
        msg.set_payload(decrmsg.get_payload())
        return (msg, valid_sig)

    def _maybe_decrypt_inline_encrypted_msg(self, origmsg, encoding,
                                            senderPubkey):
        """
        Possibly decrypt an inline OpenPGP encrypted message.

        :param origmsg: The original, possibly encrypted message.
        :type origmsg: Message
        :param encoding: The encoding of the email message.
        :type encoding: str
        :param senderPubkey: The key of the sender of the message.
        :type senderPubkey: OpenPGPKey

        :return: A tuple containing a decrypted message and
                 a bool indicating whether the signature is valid.
        :rtype: (Message, bool)
        """
        log.msg('maybe decrypting inline encrypted msg')
        # serialize the original message
        buf = StringIO()
        g = Generator(buf)
        g.flatten(origmsg)
        data = buf.getvalue()
        # handle exactly one inline PGP message
        valid_sig = False
        if PGP_BEGIN in data:
            begin = data.find(PGP_BEGIN)
            end = data.find(PGP_END)
            pgp_message = data[begin:end + len(PGP_END)]
            try:
                decrdata, valid_sig = self._decrypt_and_verify_data(
                    pgp_message, senderPubkey)
                # replace encrypted by decrypted content
                data = data.replace(pgp_message, decrdata)
            except keymanager_errors.DecryptError:
                logger.warning('Failed to decrypt potential inline encrypted '
                               'message. Storing message as is...')

        # if message is not encrypted, return raw data
        if isinstance(data, unicode):
            data = data.encode(encoding, 'replace')
        return (self._parser.parsestr(data), valid_sig)

    def _decrypt_and_verify_data(self, data, senderPubkey):
        """
        Decrypt C{data} using our private key and attempt to verify a
        signature using C{senderPubkey}.

        :param data: The text to be decrypted.
        :type data: unicode
        :param senderPubkey: The public key of the sender of the message.
        :type senderPubkey: OpenPGPKey

        :return: The decrypted data and a boolean stating whether the
                 signature could be verified.
        :rtype: (str, bool)

        :raise DecryptError: Raised if failed to decrypt.
        """
        log.msg('decrypting and verifying data')
        valid_sig = False
        try:
            decrdata = self._keymanager.decrypt(
                data, self._pkey,
                verify=senderPubkey)
            if senderPubkey is not None:
                valid_sig = True
        except keymanager_errors.InvalidSignature:
            decrdata = self._keymanager.decrypt(
                data, self._pkey)
        return (decrdata, valid_sig)

    def _add_message_locally(self, result):
        """
        Adds a message to local inbox and delete it from the incoming db
        in soledad.

        # XXX this comes from a gatherresult...
        :param msgtuple: a tuple consisting of a SoledadDocument
                         instance containing the incoming message
                         and data, the json-encoded, decrypted content of the
                         incoming message
        :type msgtuple: (SoledadDocument, str)
        """
        from twisted.internet import reactor
        msgtuple = first(result)

        doc, data = msgtuple
        log.msg('adding message %s to local db' % (doc.doc_id,))

        if isinstance(data, list):
            if empty(data):
                return False
            data = data[0]

        def msgSavedCallback(result):
            if not empty(result):
                leap_events.signal(IMAP_MSG_SAVED_LOCALLY)
                deferLater(reactor, 0, self._delete_incoming_message, doc)
                leap_events.signal(IMAP_MSG_DELETED_INCOMING)

        d = self._inbox.addMessage(data, flags=(self.RECENT_FLAG,),
                                   notify_on_disk=True)
        d.addCallbacks(msgSavedCallback, self._errback)

    #
    # helpers
    #

    def _msg_multipart_sanity_check(self, msg):
        """
        Performs a sanity check against a multipart encrypted msg

        :param msg: The original encrypted message.
        :type msg: Message
        """
        # sanity check
        payload = msg.get_payload()
        if len(payload) != 2:
            raise MalformedMessage(
                'Multipart/encrypted messages should have exactly 2 body '
                'parts (instead of %d).' % len(payload))
        if payload[0].get_content_type() != 'application/pgp-encrypted':
            raise MalformedMessage(
                "Multipart/encrypted messages' first body part should "
                "have content type equal to 'application/pgp-encrypted' "
                "(instead of %s)." % payload[0].get_content_type())
        if payload[1].get_content_type() != 'application/octet-stream':
            raise MalformedMessage(
                "Multipart/encrypted messages' second body part should "
                "have content type equal to 'octet-stream' (instead of "
                "%s)." % payload[1].get_content_type())

    def _is_msg(self, keys):
        """
        Checks if the keys of a dictionary match the signature
        of the document type we use for messages.

        :param keys: iterable containing the strings to match.
        :type keys: iterable of strings.
        :rtype: bool
        """
        return ENC_SCHEME_KEY in keys and ENC_JSON_KEY in keys
